from __future__ import annotations
from typing import Any
from ...fable_library.option import (value, default_arg)
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.types import Array
from ..ldcontext import LDContext
from ..ldobject import (LDNode, LDGraph)
from .comment import LDComment
from .defined_term import LDDefinedTerm
from .formal_parameter import LDFormalParameter

def _expr1517() -> TypeInfo:
    return class_type("ARCtrl.ROCrate.LDFile", None, LDFile)


class LDFile:
    @staticmethod
    def schema_type() -> str:
        return "http://schema.org/MediaObject"

    @staticmethod
    def name() -> str:
        return "http://schema.org/name"

    @staticmethod
    def comment() -> str:
        return "http://schema.org/comment"

    @staticmethod
    def disambiguating_description() -> str:
        return "http://schema.org/disambiguatingDescription"

    @staticmethod
    def usage_info() -> str:
        return "http://schema.org/usageInfo"

    @staticmethod
    def encoding_format() -> str:
        return "http://schema.org/encodingFormat"

    @staticmethod
    def pattern() -> str:
        return "http://schema.org/pattern"

    @staticmethod
    def about() -> str:
        return "http://schema.org/about"

    @staticmethod
    def example_of_work() -> str:
        return "http://schema.org/exampleOfWork"

    @staticmethod
    def try_get_name_as_string(dt: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = dt.TryGetPropertyAsSingleton(LDFile.name(), context)
        (pattern_matching_result, n) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                n = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def get_name_as_string(dt: LDNode, context: LDContext | None=None) -> str:
        match_value: Any | None = dt.TryGetPropertyAsSingleton(LDFile.name(), context)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                n: str = value(match_value)
                return n

            else: 
                raise Exception(("Property of `name` of object with @id `" + dt.Id) + "` was not a string")


        else: 
            raise Exception(("Could not access property `name` of object with @id `" + dt.Id) + "`")


    @staticmethod
    def set_name_as_string(dt: LDNode, name: str, context: LDContext | None=None) -> Any:
        return dt.SetProperty(LDFile.name(), name, context)

    @staticmethod
    def get_comments(dt: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        def filter(ld_object: LDNode, context_1: LDContext | None=None) -> bool:
            return LDComment.validate(ld_object, context_1)

        return dt.GetPropertyNodes(LDFile.comment(), filter, graph, context)

    @staticmethod
    def set_comments(dt: LDNode, comment: Array[LDNode], context: LDContext | None=None) -> Any:
        return dt.SetProperty(LDFile.comment(), comment, context)

    @staticmethod
    def try_get_disambiguating_description_as_string(dt: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = dt.TryGetPropertyAsSingleton(LDFile.disambiguating_description(), context)
        (pattern_matching_result, dd) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                dd = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return dd

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_disambiguating_description_as_string(dt: LDNode, disambiguating_description: str, context: LDContext | None=None) -> Any:
        return dt.SetProperty(LDFile.disambiguating_description(), disambiguating_description, context)

    @staticmethod
    def try_get_encoding_format_as_string(dt: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = dt.TryGetPropertyAsSingleton(LDFile.encoding_format(), context)
        (pattern_matching_result, ef) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                ef = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return ef

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_encoding_format_as_string(dt: LDNode, encoding_format: str, context: LDContext | None=None) -> Any:
        return dt.SetProperty(LDFile.encoding_format(), encoding_format, context)

    @staticmethod
    def try_get_example_of_work(dt: LDNode, context: LDContext | None=None) -> Any | None:
        return dt.TryGetPropertyAsSingleton(LDFile.example_of_work(), context)

    @staticmethod
    def try_get_example_of_work_as_formal_parameter(dt: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> LDNode | None:
        match_value: LDNode | None = dt.TryGetPropertyAsSingleNode(LDFile.example_of_work(), graph, context)
        if match_value is not None:
            def _arrow1515(__unit: None=None) -> bool:
                fp: LDNode = match_value
                return LDFormalParameter.validate(fp, context)

            if _arrow1515():
                fp_1: LDNode = match_value
                return fp_1

            else: 
                raise Exception(("Property of `exampleOfWork` of object with @id `" + dt.Id) + "` was not a FormalParameter")


        else: 
            return None


    @staticmethod
    def set_example_of_work(pv: LDNode, example_of_work: LDNode, context: LDContext | None=None) -> Any:
        return pv.SetProperty(LDFile.example_of_work(), example_of_work, context)

    @staticmethod
    def try_get_usage_info_as_string(dt: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = dt.TryGetPropertyAsSingleton(LDFile.usage_info(), context)
        (pattern_matching_result, ui) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                ui = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return ui

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_usage_info_as_string(dt: LDNode, usage_info: str, context: LDContext | None=None) -> Any:
        return dt.SetProperty(LDFile.usage_info(), usage_info, context)

    @staticmethod
    def try_get_pattern_as_defined_term(dt: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> LDNode | None:
        match_value: LDNode | None = dt.TryGetPropertyAsSingleNode(LDFile.pattern(), graph, context)
        if match_value is not None:
            def _arrow1516(__unit: None=None) -> bool:
                dt_1: LDNode = match_value
                return LDDefinedTerm.validate(dt_1, context)

            if _arrow1516():
                dt_2: LDNode = match_value
                return dt_2

            else: 
                raise Exception(("Property of `pattern` of object with @id `" + dt.Id) + "` was not a DefinedTerm")


        else: 
            return None


    @staticmethod
    def set_pattern_as_defined_term(dt: LDNode, pattern: LDNode, context: LDContext | None=None) -> Any:
        return dt.SetProperty(LDFile.pattern(), pattern, context)

    @staticmethod
    def try_get_about(dt: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> LDNode | None:
        return dt.TryGetPropertyAsSingleNode(LDFile.about(), graph, context)

    @staticmethod
    def set_about(dt: LDNode, about: LDNode, context: LDContext | None=None) -> Any:
        return dt.SetProperty(LDFile.about(), about, context)

    @staticmethod
    def gen_id(name: str) -> str:
        return ("" + name) + ""

    @staticmethod
    def validate(dt: LDNode, context: LDContext | None=None) -> bool:
        return dt.HasProperty(LDFile.name(), context) if dt.HasType(LDFile.schema_type(), context) else False

    @staticmethod
    def validate_cwlparameter(dt: LDNode, context: LDContext | None=None) -> bool:
        return dt.HasProperty(LDFile.example_of_work(), context) if LDFile.validate(dt, context) else False

    @staticmethod
    def create(name: str, id: str | None=None, comments: Array[LDNode] | None=None, disambiguating_description: str | None=None, encoding_format: str | None=None, usage_info: str | None=None, context: LDContext | None=None) -> LDNode:
        dt: LDNode = LDNode(LDFile.gen_id(name) if (id is None) else id, [LDFile.schema_type()], None, context)
        dt.SetProperty(LDFile.name(), name, context)
        dt.SetOptionalProperty(LDFile.comment(), comments, context)
        dt.SetOptionalProperty(LDFile.disambiguating_description(), disambiguating_description, context)
        dt.SetOptionalProperty(LDFile.encoding_format(), encoding_format, context)
        dt.SetOptionalProperty(LDFile.usage_info(), usage_info, context)
        return dt

    @staticmethod
    def create_cwlparameter(name: str, example_of_work: LDNode, id: str | None=None, context: LDContext | None=None) -> LDNode:
        dt: LDNode = LDNode(default_arg(id, name), [LDFile.schema_type()], None, context)
        dt.SetProperty(LDFile.name(), name, context)
        dt.SetProperty(LDFile.example_of_work(), example_of_work, context)
        return dt


LDFile_reflection = _expr1517

__all__ = ["LDFile_reflection"]




from __future__ import annotations
from collections.abc import Callable
from typing import (Any, TypeVar)
from ..fable_library.array_ import map as map_3
from ..fable_library.list import (map as map_1, FSharpList)
from ..fable_library.map import (to_seq, to_list)
from ..fable_library.option import (default_arg_with, map as map_4, value as value_5)
from ..fable_library.seq import map as map_2
from ..fable_library.types import (float32 as float32_1, Array)
from ..fable_library.util import (IEnumerable_1, get_enumerator, dispose, to_enumerable)
from .types import (IEncodable, IEncoderHelpers_1)

__A_ = TypeVar("__A_")

_T1 = TypeVar("_T1")

_T2 = TypeVar("_T2")

_T3 = TypeVar("_T3")

_T4 = TypeVar("_T4")

_T5 = TypeVar("_T5")

_T6 = TypeVar("_T6")

_T7 = TypeVar("_T7")

_T8 = TypeVar("_T8")

_KEY = TypeVar("_KEY")

_VALUE = TypeVar("_VALUE")

_A = TypeVar("_A")

def float32(value: float32_1) -> IEncodable:
    class ObjectExpr76(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any], value: Any=value) -> Any:
            return helpers.encode_decimal_number(value)

    return ObjectExpr76()


def list_1(values: FSharpList[IEncodable]) -> IEncodable:
    class ObjectExpr77(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any], values: Any=values) -> Any:
            def mapping(v: IEncodable) -> __A_:
                return v.Encode(helpers)

            arg: FSharpList[__A_] = map_1(mapping, values)
            return helpers.encode_list(arg)

    return ObjectExpr77()


def seq(values: IEnumerable_1[IEncodable]) -> IEncodable:
    class ObjectExpr78(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any], values: Any=values) -> Any:
            def mapping(v: IEncodable) -> __A_:
                return v.Encode(helpers)

            arg: IEnumerable_1[__A_] = map_2(mapping, values)
            return helpers.encode_seq(arg)

    return ObjectExpr78()


def resize_array(values: Array[IEncodable]) -> IEncodable:
    class ObjectExpr79(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any], values: Any=values) -> Any:
            result: Array[__A_] = []
            enumerator: Any = get_enumerator(values)
            try: 
                while enumerator.System_Collections_IEnumerator_MoveNext():
                    v: IEncodable = enumerator.System_Collections_Generic_IEnumerator_1_get_Current()
                    (result.append(v.Encode(helpers)))

            finally: 
                dispose(enumerator)

            return helpers.encode_resize_array(result)

    return ObjectExpr79()


def dict_1(values: Any) -> IEncodable:
    values_1: IEnumerable_1[tuple[str, IEncodable]] = to_seq(values)
    class ObjectExpr80(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any], values: Any=values) -> Any:
            def mapping(tupled_arg: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg[0], tupled_arg[1].Encode(helpers))

            arg: IEnumerable_1[tuple[str, __A_]] = map_2(mapping, values_1)
            return helpers.encode_object(arg)

    return ObjectExpr80()


def tuple2(enc1: Callable[[_T1], IEncodable], enc2: Callable[[_T2], IEncodable], v1: Any, v2: Any) -> IEncodable:
    values: Array[IEncodable] = [enc1(v1), enc2(v2)]
    class ObjectExpr81(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any], enc1: Any=enc1, enc2: Any=enc2, v1: Any=v1, v2: Any=v2) -> Any:
            def mapping(v: IEncodable) -> __A_:
                return v.Encode(helpers)

            arg: Array[__A_] = map_3(mapping, values, None)
            return helpers.encode_array(arg)

    return ObjectExpr81()


def tuple3(enc1: Callable[[_T1], IEncodable], enc2: Callable[[_T2], IEncodable], enc3: Callable[[_T3], IEncodable], v1: Any, v2: Any, v3: Any) -> IEncodable:
    values: Array[IEncodable] = [enc1(v1), enc2(v2), enc3(v3)]
    class ObjectExpr82(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any], enc1: Any=enc1, enc2: Any=enc2, enc3: Any=enc3, v1: Any=v1, v2: Any=v2, v3: Any=v3) -> Any:
            def mapping(v: IEncodable) -> __A_:
                return v.Encode(helpers)

            arg: Array[__A_] = map_3(mapping, values, None)
            return helpers.encode_array(arg)

    return ObjectExpr82()


def tuple4(enc1: Callable[[_T1], IEncodable], enc2: Callable[[_T2], IEncodable], enc3: Callable[[_T3], IEncodable], enc4: Callable[[_T4], IEncodable], v1: Any, v2: Any, v3: Any, v4: Any) -> IEncodable:
    values: Array[IEncodable] = [enc1(v1), enc2(v2), enc3(v3), enc4(v4)]
    class ObjectExpr83(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any], enc1: Any=enc1, enc2: Any=enc2, enc3: Any=enc3, enc4: Any=enc4, v1: Any=v1, v2: Any=v2, v3: Any=v3, v4: Any=v4) -> Any:
            def mapping(v: IEncodable) -> __A_:
                return v.Encode(helpers)

            arg: Array[__A_] = map_3(mapping, values, None)
            return helpers.encode_array(arg)

    return ObjectExpr83()


def tuple5(enc1: Callable[[_T1], IEncodable], enc2: Callable[[_T2], IEncodable], enc3: Callable[[_T3], IEncodable], enc4: Callable[[_T4], IEncodable], enc5: Callable[[_T5], IEncodable], v1: Any, v2: Any, v3: Any, v4: Any, v5: Any) -> IEncodable:
    values: Array[IEncodable] = [enc1(v1), enc2(v2), enc3(v3), enc4(v4), enc5(v5)]
    class ObjectExpr84(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any], enc1: Any=enc1, enc2: Any=enc2, enc3: Any=enc3, enc4: Any=enc4, enc5: Any=enc5, v1: Any=v1, v2: Any=v2, v3: Any=v3, v4: Any=v4, v5: Any=v5) -> Any:
            def mapping(v: IEncodable) -> __A_:
                return v.Encode(helpers)

            arg: Array[__A_] = map_3(mapping, values, None)
            return helpers.encode_array(arg)

    return ObjectExpr84()


def tuple6(enc1: Callable[[_T1], IEncodable], enc2: Callable[[_T2], IEncodable], enc3: Callable[[_T3], IEncodable], enc4: Callable[[_T4], IEncodable], enc5: Callable[[_T5], IEncodable], enc6: Callable[[_T6], IEncodable], v1: Any, v2: Any, v3: Any, v4: Any, v5: Any, v6: Any) -> IEncodable:
    values: Array[IEncodable] = [enc1(v1), enc2(v2), enc3(v3), enc4(v4), enc5(v5), enc6(v6)]
    class ObjectExpr85(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any], enc1: Any=enc1, enc2: Any=enc2, enc3: Any=enc3, enc4: Any=enc4, enc5: Any=enc5, enc6: Any=enc6, v1: Any=v1, v2: Any=v2, v3: Any=v3, v4: Any=v4, v5: Any=v5, v6: Any=v6) -> Any:
            def mapping(v: IEncodable) -> __A_:
                return v.Encode(helpers)

            arg: Array[__A_] = map_3(mapping, values, None)
            return helpers.encode_array(arg)

    return ObjectExpr85()


def tuple7(enc1: Callable[[_T1], IEncodable], enc2: Callable[[_T2], IEncodable], enc3: Callable[[_T3], IEncodable], enc4: Callable[[_T4], IEncodable], enc5: Callable[[_T5], IEncodable], enc6: Callable[[_T6], IEncodable], enc7: Callable[[_T7], IEncodable], v1: Any, v2: Any, v3: Any, v4: Any, v5: Any, v6: Any, v7: Any) -> IEncodable:
    values: Array[IEncodable] = [enc1(v1), enc2(v2), enc3(v3), enc4(v4), enc5(v5), enc6(v6), enc7(v7)]
    class ObjectExpr86(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any], enc1: Any=enc1, enc2: Any=enc2, enc3: Any=enc3, enc4: Any=enc4, enc5: Any=enc5, enc6: Any=enc6, enc7: Any=enc7, v1: Any=v1, v2: Any=v2, v3: Any=v3, v4: Any=v4, v5: Any=v5, v6: Any=v6, v7: Any=v7) -> Any:
            def mapping(v: IEncodable) -> __A_:
                return v.Encode(helpers)

            arg: Array[__A_] = map_3(mapping, values, None)
            return helpers.encode_array(arg)

    return ObjectExpr86()


def tuple8(enc1: Callable[[_T1], IEncodable], enc2: Callable[[_T2], IEncodable], enc3: Callable[[_T3], IEncodable], enc4: Callable[[_T4], IEncodable], enc5: Callable[[_T5], IEncodable], enc6: Callable[[_T6], IEncodable], enc7: Callable[[_T7], IEncodable], enc8: Callable[[_T8], IEncodable], v1: Any, v2: Any, v3: Any, v4: Any, v5: Any, v6: Any, v7: Any, v8: Any) -> IEncodable:
    values: Array[IEncodable] = [enc1(v1), enc2(v2), enc3(v3), enc4(v4), enc5(v5), enc6(v6), enc7(v7), enc8(v8)]
    class ObjectExpr87(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any], enc1: Any=enc1, enc2: Any=enc2, enc3: Any=enc3, enc4: Any=enc4, enc5: Any=enc5, enc6: Any=enc6, enc7: Any=enc7, enc8: Any=enc8, v1: Any=v1, v2: Any=v2, v3: Any=v3, v4: Any=v4, v5: Any=v5, v6: Any=v6, v7: Any=v7, v8: Any=v8) -> Any:
            def mapping(v: IEncodable) -> __A_:
                return v.Encode(helpers)

            arg: Array[__A_] = map_3(mapping, values, None)
            return helpers.encode_array(arg)

    return ObjectExpr87()


def map(key_encoder: Callable[[_KEY], IEncodable], value_encoder: Callable[[_VALUE], IEncodable], values: Any) -> IEncodable:
    def mapping(tupled_arg: tuple[_KEY, _VALUE], key_encoder: Any=key_encoder, value_encoder: Any=value_encoder, values: Any=values) -> IEncodable:
        return tuple2(key_encoder, value_encoder, tupled_arg[0], tupled_arg[1])

    return list_1(map_1(mapping, to_list(values)))


def Enum_byte(value: Any | None=None) -> IEncodable:
    class ObjectExpr88(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any], value: Any=value) -> Any:
            return helpers.encode_unsigned_integral_number(value)

    return ObjectExpr88()


def Enum_sbyte(value: Any | None=None) -> IEncodable:
    class ObjectExpr89(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any], value: Any=value) -> Any:
            return helpers.encode_signed_integral_number(value)

    return ObjectExpr89()


def Enum_int16(value: Any | None=None) -> IEncodable:
    class ObjectExpr90(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any], value: Any=value) -> Any:
            return helpers.encode_signed_integral_number(value)

    return ObjectExpr90()


def Enum_uint16(value: Any | None=None) -> IEncodable:
    class ObjectExpr91(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any], value: Any=value) -> Any:
            return helpers.encode_unsigned_integral_number(value)

    return ObjectExpr91()


def Enum_int(value: Any | None=None) -> IEncodable:
    class ObjectExpr92(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any], value: Any=value) -> Any:
            return helpers.encode_signed_integral_number(value)

    return ObjectExpr92()


def Enum_uint32(value: Any | None=None) -> IEncodable:
    class ObjectExpr93(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any], value: Any=value) -> Any:
            return helpers.encode_unsigned_integral_number(value)

    return ObjectExpr93()


def lossy_option(encoder: Callable[[_A], IEncodable]) -> Callable[[_A | None], IEncodable]:
    def _arrow95(arg: _A | None=None, encoder: Any=encoder) -> IEncodable:
        def def_thunk(__unit: None=None) -> IEncodable:
            class ObjectExpr94(IEncodable):
                def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                    return helpers.encode_null()

            return ObjectExpr94()

        return default_arg_with(map_4(encoder, arg), def_thunk)

    return _arrow95


def lossless_option(encoder: Callable[[_A], IEncodable], value: Any | None=None) -> IEncodable:
    if value is None:
        class ObjectExpr96(IEncodable):
            def Encode(self, helpers_3: IEncoderHelpers_1[Any], encoder: Any=encoder, value: Any=value) -> Any:
                return helpers_3.encode_string("option")

        class ObjectExpr97(IEncodable):
            def Encode(self, helpers_4: IEncoderHelpers_1[Any], encoder: Any=encoder, value: Any=value) -> Any:
                return helpers_4.encode_string("none")

        values_1: IEnumerable_1[tuple[str, IEncodable]] = to_enumerable([("$type", ObjectExpr96()), ("$case", ObjectExpr97())])
        class ObjectExpr98(IEncodable):
            def Encode(self, helpers_5: IEncoderHelpers_1[Any], encoder: Any=encoder, value: Any=value) -> Any:
                def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                    return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_5))

                arg_1: IEnumerable_1[tuple[str, __A_]] = map_2(mapping_1, values_1)
                return helpers_5.encode_object(arg_1)

        return ObjectExpr98()

    else: 
        v: _A = value_5(value)
        class ObjectExpr99(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any], encoder: Any=encoder, value: Any=value) -> Any:
                return helpers.encode_string("option")

        class ObjectExpr100(IEncodable):
            def Encode(self, helpers_1: IEncoderHelpers_1[Any], encoder: Any=encoder, value: Any=value) -> Any:
                return helpers_1.encode_string("some")

        values: IEnumerable_1[tuple[str, IEncodable]] = to_enumerable([("$type", ObjectExpr99()), ("$case", ObjectExpr100()), ("$value", encoder(v))])
        class ObjectExpr101(IEncodable):
            def Encode(self, helpers_2: IEncoderHelpers_1[Any], encoder: Any=encoder, value: Any=value) -> Any:
                def mapping(tupled_arg: tuple[str, IEncodable]) -> tuple[str, __A_]:
                    return (tupled_arg[0], tupled_arg[1].Encode(helpers_2))

                arg: IEnumerable_1[tuple[str, __A_]] = map_2(mapping, values)
                return helpers_2.encode_object(arg)

        return ObjectExpr101()



__all__ = ["float32", "list_1", "seq", "resize_array", "dict_1", "tuple2", "tuple3", "tuple4", "tuple5", "tuple6", "tuple7", "tuple8", "map", "Enum_byte", "Enum_sbyte", "Enum_int16", "Enum_uint16", "Enum_int", "Enum_uint32", "lossy_option", "lossless_option"]




from __future__ import annotations

import time as _time
from datetime import datetime, timedelta, tzinfo


class LocalTimezone(tzinfo):
    def utcoffset(self, dt: datetime | None = None):
        # Return offset of local timezone from UTC
        if dt is not None and self._isdst(dt):
            return timedelta(seconds=-_time.altzone)
        else:
            return timedelta(seconds=-_time.timezone)

    def dst(self, dt: datetime | None = None):
        # Return the daylight saving time (DST) adjustment
        if dt is not None and self._isdst(dt):
            return timedelta(seconds=-_time.altzone)
        else:
            return timedelta(seconds=-_time.timezone)

    def tzname(self, dt: datetime | None = None) -> str:
        # Return the name of the time zone
        return "localtime"

    def _isdst(self, dt: datetime):
        # Determine whether or not DST is in effect
        tt = (dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second, dt.weekday(), 0, 0)
        stamp = _time.mktime(tt)
        tt = _time.localtime(stamp)
        return tt.tm_isdst > 0


local_time_zone = LocalTimezone()

__all__ = ["local_time_zone"]



from __future__ import annotations
from collections.abc import Callable
from typing import (Any, TypeVar)
from ....arctrl_core.Helper.collections_ import Dictionary_init
from ....arctrl_core.ontology_annotation import OntologyAnnotation
from ....arctrl_core.Table.arc_table import ArcTable
from ....arctrl_core.Table.composite_cell import CompositeCell
from ....arctrl_json.encode import default_spaces
from ....arctrl_json.string_table import (decoder as decoder_3, encoder as encoder_3, array_from_map as array_from_map_2)
from ....arctrl_json.Table.arc_table import (decoder as decoder_2, encoder, decoder_compressed, encoder_compressed)
from ....arctrl_json.Table.cell_table import (decoder as decoder_5, encoder as encoder_1, array_from_map)
from ....arctrl_json.Table.oatable import (decoder as decoder_4, encoder as encoder_2, array_from_map as array_from_map_1)
from ....fable_library.result import FSharpResult_2
from ....fable_library.seq import map
from ....fable_library.string_ import (to_text, printf)
from ....fable_library.types import Array
from ....fable_library.util import (to_enumerable, IEnumerable_1)
from ....thoth_json_core.decode import (object, IRequiredGetter, IGetters)
from ....thoth_json_core.types import (IEncodable, Decoder_1, IEncoderHelpers_1)
from ....thoth_json_python.decode import Decode_fromString
from ....thoth_json_python.encode import to_string

__A_ = TypeVar("__A_")

def ARCtrl_ArcTable__ArcTable_fromJsonString_Static_Z721C83C5(s: str) -> ArcTable:
    match_value: FSharpResult_2[ArcTable, str] = Decode_fromString(decoder_2, s)
    if match_value.tag == 1:
        raise Exception(to_text(printf("Error decoding string: %O"))(match_value.fields[0]))

    else: 
        return match_value.fields[0]



def ARCtrl_ArcTable__ArcTable_toJsonString_Static_71136F3F(spaces: int | None=None) -> Callable[[ArcTable], str]:
    def _arrow3842(obj: ArcTable, spaces: Any=spaces) -> str:
        value: IEncodable = encoder(obj)
        return to_string(default_spaces(spaces), value)

    return _arrow3842


def ARCtrl_ArcTable__ArcTable_ToJsonString_71136F3F(this: ArcTable, spaces: int | None=None) -> str:
    return ARCtrl_ArcTable__ArcTable_toJsonString_Static_71136F3F(spaces)(this)


def ARCtrl_ArcTable__ArcTable_fromCompressedJsonString_Static_Z721C83C5(json_string: str) -> ArcTable:
    def _arrow3844(get: IGetters, json_string: Any=json_string) -> ArcTable:
        string_table: Array[str]
        object_arg: IRequiredGetter = get.Required
        string_table = object_arg.Field("stringTable", decoder_3)
        oa_table: Array[OntologyAnnotation]
        arg_3: Decoder_1[Array[OntologyAnnotation]] = decoder_4(string_table)
        object_arg_1: IRequiredGetter = get.Required
        oa_table = object_arg_1.Field("oaTable", arg_3)
        def _arrow3843(__unit: None=None) -> Array[CompositeCell]:
            arg_5: Decoder_1[Array[CompositeCell]] = decoder_5(string_table, oa_table)
            object_arg_2: IRequiredGetter = get.Required
            return object_arg_2.Field("cellTable", arg_5)

        arg_7: Decoder_1[ArcTable] = decoder_compressed(string_table, oa_table, _arrow3843())
        object_arg_3: IRequiredGetter = get.Required
        return object_arg_3.Field("table", arg_7)

    match_value: FSharpResult_2[ArcTable, str] = Decode_fromString(object(_arrow3844), json_string)
    if match_value.tag == 1:
        raise Exception(to_text(printf("Error decoding string: %O"))(match_value.fields[0]))

    else: 
        return match_value.fields[0]



def ARCtrl_ArcTable__ArcTable_ToCompressedJsonString_71136F3F(this: ArcTable, spaces: int | None=None) -> str:
    spaces_1: int = default_spaces(spaces) or 0
    string_table: Any = Dictionary_init()
    oa_table: Any = Dictionary_init()
    cell_table: Any = Dictionary_init()
    arc_table: IEncodable = encoder_compressed(string_table, oa_table, cell_table, this)
    def _arrow3846(__unit: None=None, this: Any=this, spaces: Any=spaces) -> IEncodable:
        values: IEnumerable_1[tuple[str, IEncodable]] = to_enumerable([("cellTable", encoder_1(string_table, oa_table, array_from_map(cell_table))), ("oaTable", encoder_2(string_table, array_from_map_1(oa_table))), ("stringTable", encoder_3(array_from_map_2(string_table))), ("table", arc_table)])
        class ObjectExpr3845(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                def mapping(tupled_arg: tuple[str, IEncodable]) -> tuple[str, __A_]:
                    return (tupled_arg[0], tupled_arg[1].Encode(helpers))

                arg: IEnumerable_1[tuple[str, __A_]] = map(mapping, values)
                return helpers.encode_object(arg)

        return ObjectExpr3845()

    return to_string(spaces_1, _arrow3846())


def ARCtrl_ArcTable__ArcTable_toCompressedJsonString_Static_71136F3F(spaces: int | None=None) -> Callable[[ArcTable], str]:
    def _arrow3847(obj: ArcTable, spaces: Any=spaces) -> str:
        return ARCtrl_ArcTable__ArcTable_ToCompressedJsonString_71136F3F(obj, spaces)

    return _arrow3847


__all__ = ["ARCtrl_ArcTable__ArcTable_fromJsonString_Static_Z721C83C5", "ARCtrl_ArcTable__ArcTable_toJsonString_Static_71136F3F", "ARCtrl_ArcTable__ArcTable_ToJsonString_71136F3F", "ARCtrl_ArcTable__ArcTable_fromCompressedJsonString_Static_Z721C83C5", "ARCtrl_ArcTable__ArcTable_ToCompressedJsonString_71136F3F", "ARCtrl_ArcTable__ArcTable_toCompressedJsonString_Static_71136F3F"]




from __future__ import annotations
from typing import Any
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.types import Array
from ..helper import clean
from ..ldcontext import LDContext
from ..ldobject import LDNode
from .create_action import LDCreateAction
from .lab_process import LDLabProcess

def _expr1540() -> TypeInfo:
    return class_type("ARCtrl.ROCrate.LDWorkflowInvocation", None, LDWorkflowInvocation)


class LDWorkflowInvocation:
    @staticmethod
    def schema_type() -> Array[str]:
        return [LDCreateAction.schema_type(), LDLabProcess.schema_type()]

    @staticmethod
    def gen_id(name: str, run_name: str | None=None) -> str:
        return clean(("#WorkflowInvocation_" + name) + "") if (run_name is None) else clean(((("#WorkflowInvocation_R_" + run_name) + "_") + name) + "")

    @staticmethod
    def validate(wp: LDNode, context: LDContext | None=None) -> bool:
        return LDLabProcess.validate(wp, context) if LDCreateAction.validate(wp, context) else False

    @staticmethod
    def create(name: str, instrument: LDNode, objects: Array[LDNode] | None=None, results: Array[LDNode] | None=None, description: str | None=None, agents: Array[LDNode] | None=None, id: str | None=None, end_time: Any | None=None, disambiguating_descriptions: Array[str] | None=None, executes_lab_protocol: LDNode | None=None, parameter_values: Array[LDNode] | None=None, context: LDContext | None=None) -> LDNode:
        at: Array[str] = ["WorkflowInvocation"]
        ca: LDNode = LDNode(LDWorkflowInvocation.gen_id(name) if (id is None) else id, LDWorkflowInvocation.schema_type(), at, context)
        ca.SetProperty(LDCreateAction.name(), name, context)
        ca.SetOptionalProperty(LDCreateAction.object_(), objects, context)
        ca.SetOptionalProperty(LDCreateAction.result(), results, context)
        ca.SetProperty(LDCreateAction.instrument(), instrument, context)
        ca.SetOptionalProperty(LDCreateAction.agent(), agents, context)
        ca.SetOptionalProperty(LDCreateAction.description(), description, context)
        ca.SetOptionalProperty(LDCreateAction.end_time(), end_time, context)
        ca.SetOptionalProperty(LDCreateAction.disambiguating_description(), disambiguating_descriptions, context)
        ca.SetOptionalProperty(LDLabProcess.executes_lab_protocol(), executes_lab_protocol, context)
        ca.SetOptionalProperty(LDLabProcess.parameter_value(), parameter_values, context)
        return ca


LDWorkflowInvocation_reflection = _expr1540

__all__ = ["LDWorkflowInvocation_reflection"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ....arctrl_core.Process.protocol import Protocol
from ....arctrl_json.encode import default_spaces
from ....arctrl_json.Process.protocol import (ISAJson_decoder, ISAJson_encoder, ROCrate_decoder, ROCrate_encoder)
from ....fable_library.option import default_arg
from ....fable_library.result import FSharpResult_2
from ....fable_library.string_ import (to_text, printf)
from ....thoth_json_core.types import IEncodable
from ....thoth_json_python.decode import Decode_fromString
from ....thoth_json_python.encode import to_string

def ARCtrl_Process_Protocol__Protocol_fromISAJsonString_Static_Z721C83C5(s: str) -> Protocol:
    match_value: FSharpResult_2[Protocol, str] = Decode_fromString(ISAJson_decoder, s)
    if match_value.tag == 1:
        raise Exception(to_text(printf("Error decoding string: %O"))(match_value.fields[0]))

    else: 
        return match_value.fields[0]



def ARCtrl_Process_Protocol__Protocol_toISAJsonString_Static_Z3B036AA(spaces: int | None=None, use_idreferencing: bool | None=None) -> Callable[[Protocol], str]:
    id_map: Any | None = dict([]) if default_arg(use_idreferencing, False) else None
    def _arrow3811(f: Protocol, spaces: Any=spaces, use_idreferencing: Any=use_idreferencing) -> str:
        value: IEncodable = ISAJson_encoder(None, None, None, id_map, f)
        return to_string(default_spaces(spaces), value)

    return _arrow3811


def ARCtrl_Process_Protocol__Protocol_ToISAJsonString_71136F3F(this: Protocol, spaces: int | None=None) -> str:
    return ARCtrl_Process_Protocol__Protocol_toISAJsonString_Static_Z3B036AA(spaces)(this)


def ARCtrl_Process_Protocol__Protocol_fromROCrateString_Static_Z721C83C5(s: str) -> Protocol:
    match_value: FSharpResult_2[Protocol, str] = Decode_fromString(ROCrate_decoder, s)
    if match_value.tag == 1:
        raise Exception(to_text(printf("Error decoding string: %O"))(match_value.fields[0]))

    else: 
        return match_value.fields[0]



def ARCtrl_Process_Protocol__Protocol_toROCrateString_Static_Z482224B9(study_name: str | None=None, assay_name: str | None=None, process_name: str | None=None, spaces: int | None=None) -> Callable[[Protocol], str]:
    def _arrow3812(f: Protocol, study_name: Any=study_name, assay_name: Any=assay_name, process_name: Any=process_name, spaces: Any=spaces) -> str:
        value: IEncodable = ROCrate_encoder(study_name, assay_name, process_name, f)
        return to_string(default_spaces(spaces), value)

    return _arrow3812


def ARCtrl_Process_Protocol__Protocol_ToROCrateString_Z482224B9(this: Protocol, study_name: str | None=None, assay_name: str | None=None, process_name: str | None=None, spaces: int | None=None) -> str:
    return ARCtrl_Process_Protocol__Protocol_toROCrateString_Static_Z482224B9(study_name, assay_name, process_name, spaces)(this)


__all__ = ["ARCtrl_Process_Protocol__Protocol_fromISAJsonString_Static_Z721C83C5", "ARCtrl_Process_Protocol__Protocol_toISAJsonString_Static_Z3B036AA", "ARCtrl_Process_Protocol__Protocol_ToISAJsonString_71136F3F", "ARCtrl_Process_Protocol__Protocol_fromROCrateString_Static_Z721C83C5", "ARCtrl_Process_Protocol__Protocol_toROCrateString_Static_Z482224B9", "ARCtrl_Process_Protocol__Protocol_ToROCrateString_Z482224B9"]




from __future__ import annotations
from collections.abc import Callable
from typing import (Any, TypeVar)
from ...arctrl_core.Process.material_type import MaterialType
from ...fable_library.result import FSharpResult_2
from ...thoth_json_core.decode import string
from ...thoth_json_core.types import (IEncodable, IEncoderHelpers_1, Decoder_1, ErrorReason_1, IDecoderHelpers_1)

__A_ = TypeVar("__A_")

def ROCrate_encoder(value: MaterialType) -> IEncodable:
    if value.tag == 1:
        class ObjectExpr2451(IEncodable):
            def Encode(self, helpers_1: IEncoderHelpers_1[Any], value: Any=value) -> Any:
                return helpers_1.encode_string("Labeled Extract Name")

        return ObjectExpr2451()

    else: 
        class ObjectExpr2452(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any], value: Any=value) -> Any:
                return helpers.encode_string("Extract Name")

        return ObjectExpr2452()



class ObjectExpr2453(Decoder_1[MaterialType]):
    def Decode(self, s: IDecoderHelpers_1[Any], json: Any) -> FSharpResult_2[MaterialType, tuple[str, ErrorReason_1[__A_]]]:
        match_value: FSharpResult_2[str, tuple[str, ErrorReason_1[__A_]]] = string.Decode(s, json)
        if match_value.tag == 1:
            return FSharpResult_2(1, match_value.fields[0])

        elif match_value.fields[0] == "Extract Name":
            return FSharpResult_2(0, MaterialType(0))

        elif match_value.fields[0] == "Labeled Extract Name":
            return FSharpResult_2(0, MaterialType(1))

        else: 
            s_1: str = match_value.fields[0]
            return FSharpResult_2(1, (("Could not parse " + s_1) + "No other value than \"Extract Name\" or \"Labeled Extract Name\" allowed for materialtype", ErrorReason_1(0, s_1, json)))



ROCrate_decoder: Decoder_1[MaterialType] = ObjectExpr2453()

ISAJson_encoder: Callable[[MaterialType], IEncodable] = ROCrate_encoder

ISAJson_decoder: Decoder_1[MaterialType] = ROCrate_decoder

__all__ = ["ROCrate_encoder", "ROCrate_decoder", "ISAJson_encoder", "ISAJson_decoder"]




from __future__ import annotations
from collections.abc import Callable
from typing import (Any, Generic, TypeVar)
from ...fable_library.list import (reduce, map, FSharpList, exists, pick, empty)
from ...fable_library.reflection import (TypeInfo, string_type, class_type, union_type, list_type, int32_type, obj_type, tuple_type)
from ...fable_library.string_ import (to_console, printf)
from ...fable_library.types import (Array, Union)
from ..Cells.fs_cell import DataType_reflection

_T = TypeVar("_T")

def _expr289() -> TypeInfo:
    return union_type("FsSpreadsheet.DSL.Message", [], Message, lambda: [[("Item", string_type)], [("Item", class_type("System.Exception"))]])


class Message(Union):
    def __init__(self, tag: int, *fields: Any) -> None:
        super().__init__()
        self.tag: int = tag or 0
        self.fields: Array[Any] = list(fields)

    @staticmethod
    def cases() -> list[str]:
        return ["Text", "Exception"]


Message_reflection = _expr289

def Message_message_Z721C83C5(s: str) -> Message:
    return Message(0, s)


def Message_message_2BC701FD(e: Any | None=None) -> Message:
    return Message(1, e)


def Message__MapText_11D407F6(this: Message, m: Callable[[str], str]) -> Message:
    if this.tag == 1:
        return this

    else: 
        return Message(0, m(this.fields[0]))



def Message__AsString(this: Message) -> str:
    if this.tag == 1:
        return str(this.fields[0])

    else: 
        return this.fields[0]



def Message__TryText(this: Message) -> str | None:
    if this.tag == 0:
        return this.fields[0]

    else: 
        return None



def Message__TryException(this: Message) -> Exception | None:
    if this.tag == 1:
        return this.fields[0]

    else: 
        return None



def Message__get_IsTxt(this: Message) -> bool:
    if this.tag == 0:
        return True

    else: 
        return False



def Message__get_IsExc(this: Message) -> bool:
    if this.tag == 0:
        return True

    else: 
        return False



def Messages_format(ms: FSharpList[Message]) -> str:
    def reduction(a: str, b: str, ms: Any=ms) -> str:
        return (a + ";") + b

    def mapping(m: Message, ms: Any=ms) -> str:
        return Message__AsString(m)

    return reduce(reduction, map(mapping, ms))


def Messages_fail(ms: FSharpList[Message]) -> Any:
    s: str = Messages_format(ms)
    def predicate(m: Message, ms: Any=ms) -> bool:
        return Message__get_IsExc(m)

    if exists(predicate, ms):
        to_console(printf("s"))
        def chooser(m_1: Message, ms: Any=ms) -> Exception | None:
            return Message__TryException(m_1)

        raise pick(chooser, ms)

    else: 
        raise Exception(s)



def _expr290(gen0: TypeInfo) -> TypeInfo:
    return union_type("FsSpreadsheet.DSL.SheetEntity`1", [gen0], SheetEntity_1, lambda: [[("Item1", gen0), ("Item2", list_type(Message_reflection()))], [("Item", list_type(Message_reflection()))], [("Item", list_type(Message_reflection()))]])


class SheetEntity_1(Union, Generic[_T]):
    def __init__(self, tag: int, *fields: Any) -> None:
        super().__init__()
        self.tag: int = tag or 0
        self.fields: Array[Any] = list(fields)

    @staticmethod
    def cases() -> list[str]:
        return ["Some", "NoneOptional", "NoneRequired"]


SheetEntity_1_reflection = _expr290

def SheetEntity_1_some_2B595(v: Any | None=None) -> SheetEntity_1[Any]:
    return SheetEntity_1(0, v, empty())


def SheetEntity_1__get_Messages(this: SheetEntity_1[Any]) -> FSharpList[Message]:
    if this.tag == 1:
        return this.fields[0]

    elif this.tag == 2:
        return this.fields[0]

    else: 
        return this.fields[1]



def _expr291() -> TypeInfo:
    return union_type("FsSpreadsheet.DSL.ColumnIndex", [], ColumnIndex, lambda: [[("Item", int32_type)]])


class ColumnIndex(Union):
    def __init__(self, tag: int, *fields: Any) -> None:
        super().__init__()
        self.tag: int = tag or 0
        self.fields: Array[Any] = list(fields)

    @staticmethod
    def cases() -> list[str]:
        return ["Col"]


ColumnIndex_reflection = _expr291

def ColumnIndex__get_Index(self_1: ColumnIndex) -> int:
    return self_1.fields[0]


def _expr292() -> TypeInfo:
    return union_type("FsSpreadsheet.DSL.RowIndex", [], RowIndex, lambda: [[("Item", int32_type)]])


class RowIndex(Union):
    def __init__(self, tag: int, *fields: Any) -> None:
        super().__init__()
        self.tag: int = tag or 0
        self.fields: Array[Any] = list(fields)

    @staticmethod
    def cases() -> list[str]:
        return ["Row"]


RowIndex_reflection = _expr292

def RowIndex__get_Index(self_1: RowIndex) -> int:
    return self_1.fields[0]


def _expr293() -> TypeInfo:
    return union_type("FsSpreadsheet.DSL.ColumnElement", [], ColumnElement, lambda: [[("Item1", RowIndex_reflection()), ("Item2", tuple_type(DataType_reflection(), obj_type))], [("Item", tuple_type(DataType_reflection(), obj_type))]])


class ColumnElement(Union):
    def __init__(self, tag: int, *fields: Any) -> None:
        super().__init__()
        self.tag: int = tag or 0
        self.fields: Array[Any] = list(fields)

    @staticmethod
    def cases() -> list[str]:
        return ["IndexedCell", "UnindexedCell"]


ColumnElement_reflection = _expr293

def _expr294() -> TypeInfo:
    return union_type("FsSpreadsheet.DSL.RowElement", [], RowElement, lambda: [[("Item1", ColumnIndex_reflection()), ("Item2", tuple_type(DataType_reflection(), obj_type))], [("Item", tuple_type(DataType_reflection(), obj_type))]])


class RowElement(Union):
    def __init__(self, tag: int, *fields: Any) -> None:
        super().__init__()
        self.tag: int = tag or 0
        self.fields: Array[Any] = list(fields)

    @staticmethod
    def cases() -> list[str]:
        return ["IndexedCell", "UnindexedCell"]


RowElement_reflection = _expr294

def _expr295() -> TypeInfo:
    return union_type("FsSpreadsheet.DSL.TableElement", [], TableElement, lambda: [[("Item", list_type(RowElement_reflection()))], [("Item", list_type(ColumnElement_reflection()))]])


class TableElement(Union):
    def __init__(self, tag: int, *fields: Any) -> None:
        super().__init__()
        self.tag: int = tag or 0
        self.fields: Array[Any] = list(fields)

    @staticmethod
    def cases() -> list[str]:
        return ["UnindexedRow", "UnindexedColumn"]


TableElement_reflection = _expr295

def TableElement__get_IsRow(this: TableElement) -> bool:
    if this.tag == 0:
        return True

    else: 
        return False



def TableElement__get_IsColumn(this: TableElement) -> bool:
    if this.tag == 1:
        return True

    else: 
        return False



def _expr296() -> TypeInfo:
    return union_type("FsSpreadsheet.DSL.SheetElement", [], SheetElement, lambda: [[("Item1", string_type), ("Item2", list_type(TableElement_reflection()))], [("Item1", RowIndex_reflection()), ("Item2", list_type(RowElement_reflection()))], [("Item", list_type(RowElement_reflection()))], [("Item1", ColumnIndex_reflection()), ("Item2", list_type(ColumnElement_reflection()))], [("Item", list_type(ColumnElement_reflection()))], [("Item1", RowIndex_reflection()), ("Item2", ColumnIndex_reflection()), ("Item3", tuple_type(DataType_reflection(), obj_type))], [("Item", tuple_type(DataType_reflection(), obj_type))]])


class SheetElement(Union):
    def __init__(self, tag: int, *fields: Any) -> None:
        super().__init__()
        self.tag: int = tag or 0
        self.fields: Array[Any] = list(fields)

    @staticmethod
    def cases() -> list[str]:
        return ["Table", "IndexedRow", "UnindexedRow", "IndexedColumn", "UnindexedColumn", "IndexedCell", "UnindexedCell"]


SheetElement_reflection = _expr296

def _expr297() -> TypeInfo:
    return union_type("FsSpreadsheet.DSL.WorkbookElement", [], WorkbookElement, lambda: [[("Item", list_type(SheetElement_reflection()))], [("Item1", string_type), ("Item2", list_type(SheetElement_reflection()))]])


class WorkbookElement(Union):
    def __init__(self, tag: int, *fields: Any) -> None:
        super().__init__()
        self.tag: int = tag or 0
        self.fields: Array[Any] = list(fields)

    @staticmethod
    def cases() -> list[str]:
        return ["UnnamedSheet", "NamedSheet"]


WorkbookElement_reflection = _expr297

def _expr298() -> TypeInfo:
    return union_type("FsSpreadsheet.DSL.Workbook", [], Workbook, lambda: [[("Item", list_type(WorkbookElement_reflection()))]])


class Workbook(Union):
    def __init__(self, tag: int, *fields: Any) -> None:
        super().__init__()
        self.tag: int = tag or 0
        self.fields: Array[Any] = list(fields)

    @staticmethod
    def cases() -> list[str]:
        return ["Workbook"]


Workbook_reflection = _expr298

__all__ = ["Message_reflection", "Message_message_Z721C83C5", "Message_message_2BC701FD", "Message__MapText_11D407F6", "Message__AsString", "Message__TryText", "Message__TryException", "Message__get_IsTxt", "Message__get_IsExc", "Messages_format", "Messages_fail", "SheetEntity_1_reflection", "SheetEntity_1_some_2B595", "SheetEntity_1__get_Messages", "ColumnIndex_reflection", "ColumnIndex__get_Index", "RowIndex_reflection", "RowIndex__get_Index", "ColumnElement_reflection", "RowElement_reflection", "TableElement_reflection", "TableElement__get_IsRow", "TableElement__get_IsColumn", "SheetElement_reflection", "WorkbookElement_reflection", "Workbook_reflection"]




from __future__ import annotations
from collections.abc import Callable
from typing import (Any, TypeVar)
from ...arctrl_core.ontology_annotation import OntologyAnnotation
from ...arctrl_core.Process.component import (Component, Component_createAsPV, Component__get_ComponentName, Component_decomposeName_Z721C83C5)
from ...arctrl_core.value import Value
from ...fable_library.list import (choose, of_array, FSharpList)
from ...fable_library.option import map
from ...fable_library.seq import map as map_1
from ...fable_library.util import IEnumerable_1
from ...thoth_json_core.decode import (object, IOptionalGetter, IGetters)
from ...thoth_json_core.types import (IEncodable, IEncoderHelpers_1, Decoder_1)
from ..decode import Decode_uri
from ..encode import try_include
from ..ontology_annotation import (OntologyAnnotation_encoder, OntologyAnnotation_decoder, OntologyAnnotation_ISAJson_encoder, OntologyAnnotation_ISAJson_decoder)
from ..property_value import (encoder as encoder_2, decoder as decoder_2, gen_id)
from .value import (encoder as encoder_1, decoder as decoder_1)

__A_ = TypeVar("__A_")

def encoder(value: Component) -> IEncodable:
    def chooser(tupled_arg: tuple[str, IEncodable | None], value: Any=value) -> tuple[str, IEncodable] | None:
        def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
            return (tupled_arg[0], v_1)

        return map(mapping, tupled_arg[1])

    def _arrow2464(oa: OntologyAnnotation, value: Any=value) -> IEncodable:
        return OntologyAnnotation_encoder(oa)

    def _arrow2465(value_1: Value, value: Any=value) -> IEncodable:
        return encoder_1(None, value_1)

    def _arrow2466(oa_1: OntologyAnnotation, value: Any=value) -> IEncodable:
        return OntologyAnnotation_encoder(oa_1)

    values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([try_include("type", _arrow2464, value.ComponentType), try_include("value", _arrow2465, value.ComponentValue), try_include("unit", _arrow2466, value.ComponentUnit)]))
    class ObjectExpr2467(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any], value: Any=value) -> Any:
            def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers))

            arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
            return helpers.encode_object(arg)

    return ObjectExpr2467()


def _arrow2470(get: IGetters) -> Component:
    ComponentType: OntologyAnnotation | None
    object_arg: IOptionalGetter = get.Optional
    ComponentType = object_arg.Field("type", OntologyAnnotation_decoder)
    def _arrow2468(__unit: None=None) -> Value | None:
        object_arg_1: IOptionalGetter = get.Optional
        return object_arg_1.Field("value", decoder_1)

    def _arrow2469(__unit: None=None) -> OntologyAnnotation | None:
        object_arg_2: IOptionalGetter = get.Optional
        return object_arg_2.Field("unit", OntologyAnnotation_decoder)

    return Component(_arrow2468(), _arrow2469(), ComponentType)


decoder: Decoder_1[Component] = object(_arrow2470)

ROCrate_encoder: Callable[[Component], IEncodable] = encoder_2

ROCrate_decoder: Decoder_1[Component] = decoder_2(Component_createAsPV)

def ISAJson_genID(c: Component) -> str:
    return gen_id(c)


def ISAJson_encoder(id_map: Any | None, c: Component) -> IEncodable:
    def chooser(tupled_arg: tuple[str, IEncodable | None], id_map: Any=id_map, c: Any=c) -> tuple[str, IEncodable] | None:
        def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
            return (tupled_arg[0], v_1)

        return map(mapping, tupled_arg[1])

    def _arrow2474(value: str, id_map: Any=id_map, c: Any=c) -> IEncodable:
        class ObjectExpr2473(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                return helpers.encode_string(value)

        return ObjectExpr2473()

    def _arrow2475(oa: OntologyAnnotation, id_map: Any=id_map, c: Any=c) -> IEncodable:
        return OntologyAnnotation_ISAJson_encoder(id_map, oa)

    values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([try_include("componentName", _arrow2474, Component__get_ComponentName(c)), try_include("componentType", _arrow2475, c.ComponentType)]))
    class ObjectExpr2476(IEncodable):
        def Encode(self, helpers_1: IEncoderHelpers_1[Any], id_map: Any=id_map, c: Any=c) -> Any:
            def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_1))

            arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
            return helpers_1.encode_object(arg)

    return ObjectExpr2476()


def _arrow2478(get: IGetters) -> Component:
    name: str | None
    object_arg: IOptionalGetter = get.Optional
    name = object_arg.Field("componentName", Decode_uri)
    pattern_input_1: tuple[Value | None, OntologyAnnotation | None]
    if name is None:
        pattern_input_1 = (None, None)

    else: 
        pattern_input: tuple[Value, OntologyAnnotation | None] = Component_decomposeName_Z721C83C5(name)
        pattern_input_1 = (pattern_input[0], pattern_input[1])

    def _arrow2477(__unit: None=None) -> OntologyAnnotation | None:
        object_arg_1: IOptionalGetter = get.Optional
        return object_arg_1.Field("componentType", OntologyAnnotation_ISAJson_decoder)

    return Component(pattern_input_1[0], pattern_input_1[1], _arrow2477())


ISAJson_decoder: Decoder_1[Component] = object(_arrow2478)

__all__ = ["encoder", "decoder", "ROCrate_encoder", "ROCrate_decoder", "ISAJson_genID", "ISAJson_encoder", "ISAJson_decoder"]




from __future__ import annotations
from abc import abstractmethod
from collections.abc import Callable
from io import BytesIO as BytesIO_1
import openpyxl as openpyxl_1
from typing import (Any, Protocol)
from ..fable_library.reflection import (TypeInfo, union_type, class_type)
from ..fable_library.types import (Array, Union)
from ..fable_library.util import (IEnumerable_1, IEnumerable)

def Helper_writeBytes(bytes: bytearray, path: str) -> None:
    
    # Write the bytes data to the output file path using shutil
    with open(path, 'wb') as output_file:
        output_file.write(bytes)
  


def _expr70() -> TypeInfo:
    return union_type("Fable.Openpyxl.CellType", [], CellType, lambda: [[], [], [], [], [], []])


class CellType(Union):
    def __init__(self, tag: int, *fields: Any) -> None:
        super().__init__()
        self.tag: int = tag or 0
        self.fields: Array[Any] = list(fields)

    @staticmethod
    def cases() -> list[str]:
        return ["Float", "Integer", "String", "Boolean", "DateTime", "Empty"]


CellType_reflection = _expr70

def CellType_fromCellType_Z721C83C5(cell_type: str) -> CellType:
    if cell_type == "int":
        return CellType(1)

    elif cell_type == "float":
        return CellType(0)

    elif cell_type == "str":
        return CellType(2)

    elif cell_type == "bool":
        return CellType(3)

    elif cell_type == "datetime":
        return CellType(4)

    elif cell_type == "NoneType":
        return CellType(5)

    else: 
        raise Exception(("Unknown cell type of type: \'" + cell_type) + "\'")



class Cell(Protocol):
    @property
    @abstractmethod
    def cell_type(self) -> str:
        ...

    @property
    @abstractmethod
    def value(self) -> Any:
        ...

    @value.setter
    @abstractmethod
    def value(self, __arg0: Any) -> None:
        ...


class Table(Protocol):
    @property
    @abstractmethod
    def display_name(self) -> str:
        ...

    @display_name.setter
    @abstractmethod
    def display_name(self, __arg0: str) -> None:
        ...

    @property
    @abstractmethod
    def header_row_count(self) -> int:
        ...

    @header_row_count.setter
    @abstractmethod
    def header_row_count(self, __arg0: int) -> None:
        ...

    @property
    @abstractmethod
    def id(self) -> int:
        ...

    @id.setter
    @abstractmethod
    def id(self, __arg0: int) -> None:
        ...

    @property
    @abstractmethod
    def name(self) -> str:
        ...

    @name.setter
    @abstractmethod
    def name(self, __arg0: str) -> None:
        ...

    @property
    @abstractmethod
    def ref(self) -> str:
        ...

    @ref.setter
    @abstractmethod
    def ref(self, __arg0: str) -> None:
        ...


class TableMap(Protocol):
    @abstractmethod
    def Item(self, __arg0: str) -> Table:
        ...

    @abstractmethod
    def delete(self, displayName: str) -> None:
        ...

    @abstractmethod
    def items(self) -> Array[tuple[str, str]]:
        ...

    @abstractmethod
    def values(self) -> Array[Table]:
        ...


class Worksheet(Protocol):
    @abstractmethod
    def add_table(self, __arg0: Table) -> None:
        ...

    @abstractmethod
    def append(self, __arg0: Array[Any]) -> None:
        ...

    @abstractmethod
    def delete_cols(self, start_index: int, count: int) -> None:
        ...

    @abstractmethod
    def delete_rows(self, start_index: int, count: int) -> None:
        ...

    @abstractmethod
    def delete_table(self, displayName: str) -> None:
        ...

    @property
    @abstractmethod
    def columns(self) -> Array[Array[Cell]]:
        ...

    @property
    @abstractmethod
    def rows(self) -> Array[Array[Cell]]:
        ...

    @property
    @abstractmethod
    def table_count(self) -> int:
        ...

    @property
    @abstractmethod
    def tables(self) -> TableMap:
        ...

    @property
    @abstractmethod
    def title(self) -> str:
        ...

    @title.setter
    @abstractmethod
    def title(self, __arg0: str) -> None:
        ...

    @property
    @abstractmethod
    def values(self) -> Array[Array[Any]]:
        ...

    @abstractmethod
    def insert_cols(self, __arg0: int) -> None:
        ...

    @abstractmethod
    def insert_rows(self, __arg0: int) -> None:
        ...

    @abstractmethod
    def iter_cols(self, min_row: int, max_col: int, max_row: int, action: Callable[[Array[Cell]], None]) -> None:
        ...

    @abstractmethod
    def iter_rows(self, min_row: int, max_col: int, max_row: int, action: Callable[[Array[Cell]], None]) -> None:
        ...


class Workbook(IEnumerable_1, IEnumerable[Any]):
    @abstractmethod
    def Item(self, __arg0: str) -> Worksheet:
        ...

    @abstractmethod
    def copy_worksheet(self, __arg0: Worksheet) -> Worksheet:
        ...

    @abstractmethod
    def create_sheet(self, __arg0: str, position: int | None) -> Worksheet:
        ...

    @property
    @abstractmethod
    def active(self) -> Worksheet:
        ...

    @property
    @abstractmethod
    def iso_dates(self) -> bool:
        ...

    @iso_dates.setter
    @abstractmethod
    def iso_dates(self, __arg0: bool) -> None:
        ...

    @property
    @abstractmethod
    def sheetnames(self) -> Array[str]:
        ...

    @property
    @abstractmethod
    def template(self) -> bool:
        ...

    @template.setter
    @abstractmethod
    def template(self, __arg0: bool) -> None:
        ...

    @property
    @abstractmethod
    def worksheets(self) -> Array[Worksheet]:
        ...

    @worksheets.setter
    @abstractmethod
    def worksheets(self, __arg0: Array[Worksheet]) -> None:
        ...

    @abstractmethod
    def index(self, __arg0: Worksheet) -> int:
        ...

    @abstractmethod
    def remove(self, __arg0: Worksheet) -> None:
        ...


class BytesIO(Protocol):
    @abstractmethod
    def ToFile(self, path: str) -> None:
        ...

    @property
    @abstractmethod
    def x(self) -> str:
        ...

    @abstractmethod
    def getbuffer(self) -> Any:
        ...

    @abstractmethod
    def getvalue(self) -> bytearray:
        ...


class OpenPyXL(Protocol):
    @abstractmethod
    def Workbook(self) -> Workbook:
        ...


class WorkbookStatic(Protocol):
    @abstractmethod
    def create(self) -> Workbook:
        ...


class TableStatic(Protocol):
    @abstractmethod
    def create(self, displayName: str, ref: str) -> Table:
        ...


class BytesIOStatic(Protocol):
    pass

openpyxl: OpenPyXL = openpyxl_1

def _expr71() -> TypeInfo:
    return class_type("Fable.Openpyxl.Xlsx", None, Xlsx)


class Xlsx:
    ...

Xlsx_reflection = _expr71

def Xlsx_readFile_Z721C83C5(path: str) -> Workbook:
    return openpyxl.load_workbook(path)


def Xlsx_read_Z3F6BC7B1(bytes: bytearray) -> Workbook:
    return openpyxl.load_workbook(bytes)


def Xlsx_load_4E60E31B(buffer: Any=None) -> Workbook:
    return openpyxl.load_workbook(buffer)


def Xlsx_writeFile_A9E7E13(wb: Workbook, path: str) -> None:
    wb.save(path)


def Xlsx_write_Z22AEA2D8(wb: Workbook) -> bytearray:
    output: BytesIO = BytesIO_1(None)
    wb.save(output)
    return output.getvalue()


def Xlsx_writeBuffer_Z22AEA2D8(wb: Workbook) -> Any:
    output: BytesIO = BytesIO_1(None)
    wb.save(output)
    return output.getbuffer()


__all__ = ["Helper_writeBytes", "CellType_reflection", "CellType_fromCellType_Z721C83C5", "openpyxl", "Xlsx_reflection", "Xlsx_readFile_Z721C83C5", "Xlsx_read_Z3F6BC7B1", "Xlsx_load_4E60E31B", "Xlsx_writeFile_A9E7E13", "Xlsx_write_Z22AEA2D8", "Xlsx_writeBuffer_Z22AEA2D8"]




from __future__ import annotations
from ..dynamic_obj.dynamic_obj import (DynamicObj, DynamicObj_reflection)
from ..fable_library.option import default_arg
from ..fable_library.reflection import (TypeInfo, class_type)
from ..fable_library.types import Array
from .inputs import CWLInput
from .outputs import CWLOutput
from .requirements import Requirement
from .workflow_steps import WorkflowStep

def _expr519() -> TypeInfo:
    return class_type("ARCtrl.CWL.CWLWorkflowDescription", None, CWLWorkflowDescription, DynamicObj_reflection())


class CWLWorkflowDescription(DynamicObj):
    def __init__(self, steps: Array[WorkflowStep], inputs: Array[CWLInput], outputs: Array[CWLOutput], cwl_version: str | None=None, requirements: Array[Requirement] | None=None, hints: Array[Requirement] | None=None, metadata: DynamicObj | None=None) -> None:
        super().__init__()
        self._cwlVersion: str = default_arg(cwl_version, "v1.2")
        self._steps: Array[WorkflowStep] = steps
        self._inputs: Array[CWLInput] = inputs
        self._outputs: Array[CWLOutput] = outputs
        self._requirements: Array[Requirement] | None = requirements
        self._hints: Array[Requirement] | None = hints
        self._metadata: DynamicObj | None = metadata

    @property
    def CWLVersion(self, __unit: None=None) -> str:
        this: CWLWorkflowDescription = self
        return this._cwlVersion

    @CWLVersion.setter
    def CWLVersion(self, version: str) -> None:
        this: CWLWorkflowDescription = self
        this._cwlVersion = version

    @property
    def Steps(self, __unit: None=None) -> Array[WorkflowStep]:
        this: CWLWorkflowDescription = self
        return this._steps

    @Steps.setter
    def Steps(self, steps: Array[WorkflowStep]) -> None:
        this: CWLWorkflowDescription = self
        this._steps = steps

    @property
    def Inputs(self, __unit: None=None) -> Array[CWLInput]:
        this: CWLWorkflowDescription = self
        return this._inputs

    @Inputs.setter
    def Inputs(self, inputs: Array[CWLInput]) -> None:
        this: CWLWorkflowDescription = self
        this._inputs = inputs

    @property
    def Outputs(self, __unit: None=None) -> Array[CWLOutput]:
        this: CWLWorkflowDescription = self
        return this._outputs

    @Outputs.setter
    def Outputs(self, outputs: Array[CWLOutput]) -> None:
        this: CWLWorkflowDescription = self
        this._outputs = outputs

    @property
    def Requirements(self, __unit: None=None) -> Array[Requirement] | None:
        this: CWLWorkflowDescription = self
        return this._requirements

    @Requirements.setter
    def Requirements(self, requirements: Array[Requirement] | None=None) -> None:
        this: CWLWorkflowDescription = self
        this._requirements = requirements

    @property
    def Hints(self, __unit: None=None) -> Array[Requirement] | None:
        this: CWLWorkflowDescription = self
        return this._hints

    @Hints.setter
    def Hints(self, hints: Array[Requirement] | None=None) -> None:
        this: CWLWorkflowDescription = self
        this._hints = hints

    @property
    def Metadata(self, __unit: None=None) -> DynamicObj | None:
        this: CWLWorkflowDescription = self
        return this._metadata

    @Metadata.setter
    def Metadata(self, metadata: DynamicObj | None=None) -> None:
        this: CWLWorkflowDescription = self
        this._metadata = metadata


CWLWorkflowDescription_reflection = _expr519

def CWLWorkflowDescription__ctor_704BBD06(steps: Array[WorkflowStep], inputs: Array[CWLInput], outputs: Array[CWLOutput], cwl_version: str | None=None, requirements: Array[Requirement] | None=None, hints: Array[Requirement] | None=None, metadata: DynamicObj | None=None) -> CWLWorkflowDescription:
    return CWLWorkflowDescription(steps, inputs, outputs, cwl_version, requirements, hints, metadata)


__all__ = ["CWLWorkflowDescription_reflection"]




from __future__ import annotations
from typing import Any
from ...fable_library.list import (FSharpList, choose, empty)
from ...fable_library.option import (default_arg, map)
from ...fable_library.reflection import (TypeInfo, union_type)
from ...fable_library.string_ import (to_text, printf)
from ...fable_library.types import (to_string, Array, Union)
from ..data import (Data_reflection, Data)
from ..data_file import DataFile
from ..ontology_annotation import OntologyAnnotation
from .factor_value import FactorValue
from .material import (Material_reflection, Material, Material_setCharacteristicValues, Material_getUnits_ZBCDEB61, Material_create_Z66909A6D)
from .material_attribute import MaterialAttribute
from .material_attribute_value import MaterialAttributeValue
from .sample import (Sample_reflection, Sample, Sample_setCharacteristicValues, Sample_getUnits_53716792, Sample_create_62424AD2)
from .source import (Source_reflection, Source_get_empty, Source, Source_setCharacteristicValues, Source_getUnits_53E41069, Source_create_Z5CA08497)

def _expr767() -> TypeInfo:
    return union_type("ARCtrl.Process.ProcessInput", [], ProcessInput, lambda: [[("Item", Source_reflection())], [("Item", Sample_reflection())], [("Item", Data_reflection())], [("Item", Material_reflection())]])


class ProcessInput(Union):
    def __init__(self, tag: int, *fields: Any) -> None:
        super().__init__()
        self.tag: int = tag or 0
        self.fields: Array[Any] = list(fields)

    @staticmethod
    def cases() -> list[str]:
        return ["Source", "Sample", "Data", "Material"]

    def Print(self, __unit: None=None) -> str:
        this: ProcessInput = self
        return to_string(this)

    def PrintCompact(self, __unit: None=None) -> str:
        this: ProcessInput = self
        if this.tag == 0:
            arg_1: str = this.fields[0].PrintCompact()
            return to_text(printf("Source {%s}"))(arg_1)

        elif this.tag == 3:
            arg_2: str = this.fields[0].PrintCompact()
            return to_text(printf("Material {%s}"))(arg_2)

        elif this.tag == 2:
            arg_3: str = this.fields[0].PrintCompact()
            return to_text(printf("Data {%s}"))(arg_3)

        else: 
            arg: str = this.fields[0].PrintCompact()
            return to_text(printf("Sample {%s}"))(arg)



ProcessInput_reflection = _expr767

def ProcessInput__get_TryName(this: ProcessInput) -> str | None:
    if this.tag == 0:
        return this.fields[0].Name

    elif this.tag == 3:
        return this.fields[0].Name

    elif this.tag == 2:
        return this.fields[0].Name

    else: 
        return this.fields[0].Name



def ProcessInput__get_Name(this: ProcessInput) -> str:
    return default_arg(ProcessInput__get_TryName(this), "")


def ProcessInput_get_Default(__unit: None=None) -> ProcessInput:
    return ProcessInput(0, Source_get_empty())


def ProcessInput_tryGetName_5B3D5BA9(pi: ProcessInput) -> str | None:
    return ProcessInput__get_TryName(pi)


def ProcessInput_getName_5B3D5BA9(pi: ProcessInput) -> str:
    return ProcessInput__get_Name(pi)


def ProcessInput_nameEquals(name: str, pi: ProcessInput) -> bool:
    return ProcessInput__get_Name(pi) == name


def ProcessInput_isSample_5B3D5BA9(pi: ProcessInput) -> bool:
    if pi.tag == 1:
        return True

    else: 
        return False



def ProcessInput_isSource_5B3D5BA9(pi: ProcessInput) -> bool:
    if pi.tag == 0:
        return True

    else: 
        return False



def ProcessInput_isData_5B3D5BA9(pi: ProcessInput) -> bool:
    if pi.tag == 2:
        return True

    else: 
        return False



def ProcessInput_isMaterial_5B3D5BA9(pi: ProcessInput) -> bool:
    if pi.tag == 3:
        return True

    else: 
        return False



def ProcessInput__isSource(this: ProcessInput) -> bool:
    return ProcessInput_isSource_5B3D5BA9(this)


def ProcessInput__isSample(this: ProcessInput) -> bool:
    return ProcessInput_isSample_5B3D5BA9(this)


def ProcessInput__isData(this: ProcessInput) -> bool:
    return ProcessInput_isData_5B3D5BA9(this)


def ProcessInput__isMaterial(this: ProcessInput) -> bool:
    return ProcessInput_isMaterial_5B3D5BA9(this)


def ProcessInput_trySample_5B3D5BA9(pi: ProcessInput) -> Sample | None:
    if pi.tag == 1:
        return pi.fields[0]

    else: 
        return None



def ProcessInput_trySource_5B3D5BA9(pi: ProcessInput) -> Source | None:
    if pi.tag == 0:
        return pi.fields[0]

    else: 
        return None



def ProcessInput_tryData_5B3D5BA9(pi: ProcessInput) -> Data | None:
    if pi.tag == 2:
        return pi.fields[0]

    else: 
        return None



def ProcessInput_tryMaterial_5B3D5BA9(pi: ProcessInput) -> Material | None:
    if pi.tag == 3:
        return pi.fields[0]

    else: 
        return None



def ProcessInput_setCharacteristicValues(characteristics: FSharpList[MaterialAttributeValue], pi: ProcessInput) -> ProcessInput:
    if pi.tag == 0:
        return ProcessInput(0, Source_setCharacteristicValues(characteristics, pi.fields[0]))

    elif pi.tag == 3:
        return ProcessInput(3, Material_setCharacteristicValues(characteristics, pi.fields[0]))

    elif pi.tag == 2:
        return pi

    else: 
        return ProcessInput(1, Sample_setCharacteristicValues(characteristics, pi.fields[0]))



def ProcessInput_tryGetCharacteristicValues_5B3D5BA9(pi: ProcessInput) -> FSharpList[MaterialAttributeValue] | None:
    if pi.tag == 0:
        return pi.fields[0].Characteristics

    elif pi.tag == 3:
        return pi.fields[0].Characteristics

    elif pi.tag == 2:
        return None

    else: 
        return pi.fields[0].Characteristics



def ProcessInput_tryGetCharacteristics_5B3D5BA9(pi: ProcessInput) -> FSharpList[MaterialAttribute] | None:
    def mapping(list_1: FSharpList[MaterialAttributeValue], pi: Any=pi) -> FSharpList[MaterialAttribute]:
        def chooser(c: MaterialAttributeValue, list_1: Any=list_1) -> MaterialAttribute | None:
            return c.Category

        return choose(chooser, list_1)

    return map(mapping, ProcessInput_tryGetCharacteristicValues_5B3D5BA9(pi))


def ProcessInput_getCharacteristicValues_5B3D5BA9(pi: ProcessInput) -> FSharpList[MaterialAttributeValue]:
    return default_arg(ProcessInput_tryGetCharacteristicValues_5B3D5BA9(pi), empty())


def ProcessInput_getUnits_5B3D5BA9(pi: ProcessInput) -> FSharpList[OntologyAnnotation]:
    if pi.tag == 1:
        return Sample_getUnits_53716792(pi.fields[0])

    elif pi.tag == 3:
        return Material_getUnits_ZBCDEB61(pi.fields[0])

    elif pi.tag == 2:
        return empty()

    else: 
        return Source_getUnits_53E41069(pi.fields[0])



def ProcessInput_createSource_Z5E00540E(name: str, characteristics: FSharpList[MaterialAttributeValue] | None=None) -> ProcessInput:
    return ProcessInput(0, Source_create_Z5CA08497(None, name, characteristics))


def ProcessInput_createSample_Z598187B7(name: str, characteristics: FSharpList[MaterialAttributeValue] | None=None, factors: FSharpList[FactorValue] | None=None, derives_from: FSharpList[Source] | None=None) -> ProcessInput:
    return ProcessInput(1, Sample_create_62424AD2(None, name, characteristics, factors, derives_from))


def ProcessInput_createMaterial_4452CB4C(name: str, characteristics: FSharpList[MaterialAttributeValue] | None=None, derives_from: FSharpList[Material] | None=None) -> ProcessInput:
    return ProcessInput(3, Material_create_Z66909A6D(None, name, None, characteristics, derives_from))


def ProcessInput_createImageFile_Z721C83C5(name: str) -> ProcessInput:
    return ProcessInput(2, Data.create(None, name, DataFile(2)))


def ProcessInput_createRawData_Z721C83C5(name: str) -> ProcessInput:
    return ProcessInput(2, Data.create(None, name, DataFile(0)))


def ProcessInput_createDerivedData_Z721C83C5(name: str) -> ProcessInput:
    return ProcessInput(2, Data.create(None, name, DataFile(1)))


__all__ = ["ProcessInput_reflection", "ProcessInput__get_TryName", "ProcessInput__get_Name", "ProcessInput_get_Default", "ProcessInput_tryGetName_5B3D5BA9", "ProcessInput_getName_5B3D5BA9", "ProcessInput_nameEquals", "ProcessInput_isSample_5B3D5BA9", "ProcessInput_isSource_5B3D5BA9", "ProcessInput_isData_5B3D5BA9", "ProcessInput_isMaterial_5B3D5BA9", "ProcessInput__isSource", "ProcessInput__isSample", "ProcessInput__isData", "ProcessInput__isMaterial", "ProcessInput_trySample_5B3D5BA9", "ProcessInput_trySource_5B3D5BA9", "ProcessInput_tryData_5B3D5BA9", "ProcessInput_tryMaterial_5B3D5BA9", "ProcessInput_setCharacteristicValues", "ProcessInput_tryGetCharacteristicValues_5B3D5BA9", "ProcessInput_tryGetCharacteristics_5B3D5BA9", "ProcessInput_getCharacteristicValues_5B3D5BA9", "ProcessInput_getUnits_5B3D5BA9", "ProcessInput_createSource_Z5E00540E", "ProcessInput_createSample_Z598187B7", "ProcessInput_createMaterial_4452CB4C", "ProcessInput_createImageFile_Z721C83C5", "ProcessInput_createRawData_Z721C83C5", "ProcessInput_createDerivedData_Z721C83C5"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ..arctrl_core.arc_types import ArcAssay
from ..arctrl_core.comment import Remark
from ..arctrl_core.data_map import DataMap
from ..arctrl_core.Helper.identifier import create_missing_identifier
from ..arctrl_core.person import Person
from ..arctrl_core.Table.arc_table import ArcTable
from ..fable_library.list import (FSharpList, is_empty, empty, singleton as singleton_1, of_seq)
from ..fable_library.option import (default_arg, to_array)
from ..fable_library.seq import (exists, head, try_head, delay, append, singleton, iterate_indexed, map, try_find, choose, try_pick, is_empty as is_empty_1, iterate)
from ..fable_library.string_ import (starts_with_exact, to_fail, printf, to_console)
from ..fable_library.types import Array
from ..fable_library.util import (IEnumerable_1, get_enumerator, IEnumerator, to_enumerable)
from ..fs_spreadsheet.fs_row import FsRow
from ..fs_spreadsheet.fs_workbook import FsWorkbook
from ..fs_spreadsheet.fs_worksheet import FsWorksheet
from .AnnotationTable.arc_table import (try_from_fs_worksheet, to_fs_worksheet as to_fs_worksheet_1)
from .DataMapTable.data_map_table import (try_from_fs_worksheet as try_from_fs_worksheet_1, to_fs_worksheet)
from .Metadata.assays import (from_rows, to_rows)
from .Metadata.contacts import (from_rows as from_rows_1, to_rows as to_rows_1)
from .Metadata.sparse_table import (SparseRowModule_tryGetValueAt, SparseRowModule_fromValues, SparseRowModule_writeToSheet, SparseRowModule_fromFsRow, SparseRowModule_getAllValues, SparseRowModule_fromAllValues)

def ArcAssay_fromRows(rows: IEnumerable_1[IEnumerable_1[tuple[int, str]]]) -> ArcAssay:
    def predicate(row: IEnumerable_1[tuple[int, str]], rows: Any=rows) -> bool:
        return starts_with_exact(head(row)[1], "Assay")

    pattern_input: tuple[str | None, str | None] = (("Assay", "Assay Person")) if exists(predicate, rows) else ((None, None))
    en: IEnumerator[IEnumerable_1[tuple[int, str]]] = get_enumerator(rows)
    def loop(last_row_mut: str | None, assays_mut: FSharpList[ArcAssay], contacts_mut: FSharpList[Person], row_number_mut: int, rows: Any=rows) -> ArcAssay:
        while True:
            (last_row, assays, contacts, row_number) = (last_row_mut, assays_mut, contacts_mut, row_number_mut)
            (pattern_matching_result,) = (None,)
            if last_row is not None:
                def _arrow3634(__unit: None=None, last_row: Any=last_row, assays: Any=assays, contacts: Any=contacts, row_number: Any=row_number) -> bool:
                    prefix: str = last_row
                    return True if (prefix == "ASSAY") else (prefix == "ASSAY METADATA")

                if _arrow3634():
                    pattern_matching_result = 0

                elif last_row == "ASSAY PERFORMERS":
                    pattern_matching_result = 1

                else: 
                    pattern_matching_result = 2


            else: 
                pattern_matching_result = 2

            if pattern_matching_result == 0:
                pattern_input_1: tuple[str | None, int, FSharpList[Remark], FSharpList[ArcAssay]] = from_rows(pattern_input[0], row_number + 1, en)
                last_row_mut = pattern_input_1[0]
                assays_mut = pattern_input_1[3]
                contacts_mut = contacts
                row_number_mut = pattern_input_1[1]
                continue

            elif pattern_matching_result == 1:
                pattern_input_2: tuple[str | None, int, FSharpList[Remark], FSharpList[Person]] = from_rows_1(pattern_input[1], row_number + 1, en)
                last_row_mut = pattern_input_2[0]
                assays_mut = assays
                contacts_mut = pattern_input_2[3]
                row_number_mut = pattern_input_2[1]
                continue

            elif pattern_matching_result == 2:
                (pattern_matching_result_1, assays_2, contacts_2) = (None, None, None)
                if is_empty(assays):
                    if is_empty(contacts):
                        pattern_matching_result_1 = 0

                    else: 
                        pattern_matching_result_1 = 1
                        assays_2 = assays
                        contacts_2 = contacts


                else: 
                    pattern_matching_result_1 = 1
                    assays_2 = assays
                    contacts_2 = contacts

                if pattern_matching_result_1 == 0:
                    return ArcAssay.create(create_missing_identifier())

                elif pattern_matching_result_1 == 1:
                    performers: Array[Person] = list(contacts_2)
                    assay: ArcAssay = default_arg(try_head(assays_2), ArcAssay.create(create_missing_identifier()))
                    return ArcAssay.set_performers(performers, assay)


            break

    if en.System_Collections_IEnumerator_MoveNext():
        return loop(SparseRowModule_tryGetValueAt(0, en.System_Collections_Generic_IEnumerator_1_get_Current()), empty(), empty(), 1)

    else: 
        raise Exception("empty assay metadata sheet")



def ArcAssay_toRows(assay: ArcAssay) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
    def _arrow3638(__unit: None=None, assay: Any=assay) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
        def _arrow3637(__unit: None=None) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
            def _arrow3636(__unit: None=None) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
                def _arrow3635(__unit: None=None) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
                    return to_rows_1("Assay Person", of_seq(assay.Performers))

                return append(singleton(SparseRowModule_fromValues(to_enumerable(["ASSAY PERFORMERS"]))), delay(_arrow3635))

            return append(to_rows("Assay", singleton_1(assay)), delay(_arrow3636))

        return append(singleton(SparseRowModule_fromValues(to_enumerable(["ASSAY"]))), delay(_arrow3637))

    return delay(_arrow3638)


def ArcAssay_toMetadataSheet(assay: ArcAssay) -> FsWorksheet:
    sheet: FsWorksheet = FsWorksheet("isa_assay")
    def action(row_i: int, r: IEnumerable_1[tuple[int, str]], assay: Any=assay) -> None:
        SparseRowModule_writeToSheet(row_i + 1, r, sheet)

    iterate_indexed(action, ArcAssay_toRows(assay))
    return sheet


def ArcAssay_fromMetadataSheet(sheet: FsWorksheet) -> ArcAssay:
    try: 
        rows: IEnumerable_1[IEnumerable_1[tuple[int, str]]] = map(SparseRowModule_fromFsRow, sheet.Rows)
        def predicate(row: IEnumerable_1[tuple[int, str]]) -> bool:
            return starts_with_exact(head(row)[1], "Assay")

        has_prefix: bool = exists(predicate, rows)
        return ArcAssay_fromRows(rows)

    except Exception as err:
        arg: str = str(err)
        return to_fail(printf("Failed while parsing metadatasheet: %s"))(arg)



def ArcAssay_toMetadataCollection(assay: ArcAssay) -> IEnumerable_1[IEnumerable_1[str | None]]:
    def mapping(row: IEnumerable_1[tuple[int, str]], assay: Any=assay) -> IEnumerable_1[str | None]:
        return SparseRowModule_getAllValues(row)

    return map(mapping, ArcAssay_toRows(assay))


def ArcAssay_fromMetadataCollection(collection: IEnumerable_1[IEnumerable_1[str | None]]) -> ArcAssay:
    try: 
        return ArcAssay_fromRows(map(SparseRowModule_fromAllValues, collection))

    except Exception as err:
        arg: str = str(err)
        return to_fail(printf("Failed while parsing metadatasheet: %s"))(arg)



def ArcAssay_isMetadataSheetName(name: str) -> bool:
    if name == "isa_assay":
        return True

    else: 
        return name == "Assay"



def ArcAssay_isMetadataSheet(sheet: FsWorksheet) -> bool:
    return ArcAssay_isMetadataSheetName(sheet.Name)


def ArcAssay_tryGetMetadataSheet(doc: FsWorkbook) -> FsWorksheet | None:
    def predicate(sheet: FsWorksheet, doc: Any=doc) -> bool:
        return ArcAssay_isMetadataSheet(sheet)

    return try_find(predicate, doc.GetWorksheets())


def ARCtrl_ArcAssay__ArcAssay_fromFsWorkbook_Static_32154C9D(doc: FsWorkbook) -> ArcAssay:
    try: 
        assay_metadata: ArcAssay
        match_value: FsWorksheet | None = ArcAssay_tryGetMetadataSheet(doc)
        if match_value is None:
            to_console(printf("Cannot retrieve metadata: Assay file does not contain \"%s\" or \"%s\" sheet."))("isa_assay")("Assay")
            assay_metadata = ArcAssay.create(create_missing_identifier())

        else: 
            assay_metadata = ArcAssay_fromMetadataSheet(match_value)

        sheets: Array[FsWorksheet] = doc.GetWorksheets()
        annotation_tables: IEnumerable_1[ArcTable] = choose(try_from_fs_worksheet, sheets)
        datamap_sheet: DataMap | None = try_pick(try_from_fs_worksheet_1, sheets)
        if not is_empty_1(annotation_tables):
            assay_metadata.Tables = list(annotation_tables)

        assay_metadata.DataMap = datamap_sheet
        return assay_metadata

    except Exception as err:
        arg_2: str = str(err)
        return to_fail(printf("Could not parse assay: \n%s"))(arg_2)



def ARCtrl_ArcAssay__ArcAssay_toFsWorkbook_Static_Z2508BE4F(assay: ArcAssay, datamap_sheet: bool | None=None) -> FsWorkbook:
    datamap_sheet_1: bool = default_arg(datamap_sheet, True)
    doc: FsWorkbook = FsWorkbook()
    metadata_sheet: FsWorksheet = ArcAssay_toMetadataSheet(assay)
    doc.AddWorksheet(metadata_sheet)
    if datamap_sheet_1:
        def action(arg: DataMap, assay: Any=assay, datamap_sheet: Any=datamap_sheet) -> None:
            sheet: FsWorksheet = to_fs_worksheet(arg)
            doc.AddWorksheet(sheet)

        iterate(action, to_array(assay.DataMap))

    def action_1(i: int, arg_1: ArcTable, assay: Any=assay, datamap_sheet: Any=datamap_sheet) -> None:
        sheet_1: FsWorksheet = to_fs_worksheet_1(i, arg_1)
        doc.AddWorksheet(sheet_1)

    iterate_indexed(action_1, assay.Tables)
    return doc


def ARCtrl_ArcAssay__ArcAssay_ToFsWorkbook_6FCE9E49(this: ArcAssay, datamap_sheet: bool | None=None) -> FsWorkbook:
    return ARCtrl_ArcAssay__ArcAssay_toFsWorkbook_Static_Z2508BE4F(this, datamap_sheet)


__all__ = ["ArcAssay_fromRows", "ArcAssay_toRows", "ArcAssay_toMetadataSheet", "ArcAssay_fromMetadataSheet", "ArcAssay_toMetadataCollection", "ArcAssay_fromMetadataCollection", "ArcAssay_isMetadataSheetName", "ArcAssay_isMetadataSheet", "ArcAssay_tryGetMetadataSheet", "ARCtrl_ArcAssay__ArcAssay_fromFsWorkbook_Static_32154C9D", "ARCtrl_ArcAssay__ArcAssay_toFsWorkbook_Static_Z2508BE4F", "ARCtrl_ArcAssay__ArcAssay_ToFsWorkbook_6FCE9E49"]




import math
from math import copysign
from typing import Any

from .types import FSharpRef


def abs(x: float) -> float:
    return -x if x < 0 else x


def sign(x: float) -> float:
    return -1 if x < 0 else 1 if x > 0 else 0


def max(x: float, y: float) -> float:
    return x if x > y else y


def min(x: float, y: float) -> float:
    return x if x < y else y


def divide(x: float, y: float) -> float:
    """Divide two numbers.

    Returns nan if y is 0.
    """
    if y == 0:
        if x == 0:
            return float("nan")
        return float("inf") if copysign(1, x) == copysign(1, y) else float("-inf")

    return x / y


def log(x: float) -> float:
    if x == 0:
        return float("-inf")
    elif x < 0:
        return float("nan")
    return math.log(x)


def sqrt(x: float) -> float:
    try:
        return math.sqrt(x)
    except ValueError:
        return float("nan")


def is_positive_inf(value: float) -> bool:
    return math.isinf(value) and value > 0


def is_negative_inf(value: float) -> bool:
    return math.isinf(value) and value < 0


def parse(value: Any) -> float:
    try:
        return float(value)
    except Exception:
        raise ValueError(f"The input string {value} was not in a correct format.")


def try_parse(string: str, def_value: FSharpRef[float]) -> bool:
    try:
        def_value.contents = parse(string)
        return True
    except Exception:
        return False


__all__ = [
    "abs",
    "sign",
    "max",
    "min",
    "parse",
    "try_parse",
    "divide",
    "log",
    "sqrt",
    "is_negative_inf",
    "is_positive_inf",
]



from typing import (Any, TypeVar)
from ..fable_library.array_ import fold
from ..fable_library.date import (year, month, day, hour, minute, second)
from ..fable_library.option import value as value_1
from ..fable_library.seq import fold as fold_1
from ..fable_library.types import Array
from ..fable_library.util import (identity_hash, number_hash, IEnumerable_1)

__A = TypeVar("__A")

_A = TypeVar("_A")

def merge_hashes(hash1: int, hash2: int) -> int:
    return ((-1640531527 + hash2) + (hash1 << 6)) + (hash1 >> 2)


def hash_date_time(dt: Any) -> int:
    acc: int = 0
    acc = merge_hashes(acc, year(dt)) or 0
    acc = merge_hashes(acc, month(dt)) or 0
    acc = merge_hashes(acc, day(dt)) or 0
    acc = merge_hashes(acc, hour(dt)) or 0
    acc = merge_hashes(acc, minute(dt)) or 0
    acc = merge_hashes(acc, second(dt)) or 0
    return acc


def hash_1(obj: Any | None=None) -> int:
    copy_of_struct: __A = obj
    return identity_hash(copy_of_struct)


def box_hash_option(a: Any | None=None) -> Any:
    def _arrow74(__unit: None=None, a: Any=a) -> int:
        copy_of_struct: _A = value_1(a)
        return identity_hash(copy_of_struct)

    def _arrow75(__unit: None=None, a: Any=a) -> int:
        copy_of_struct_1: int = 0
        return number_hash(copy_of_struct_1)

    return _arrow74() if (a is not None) else _arrow75()


def box_hash_array(a: Array[Any]) -> Any:
    def folder(acc: int, o: _A, a: Any=a) -> int:
        return merge_hashes(acc, hash_1(o))

    return fold(folder, 0, a)


def box_hash_seq(a: IEnumerable_1[Any]) -> Any:
    def folder(acc: int, o: _A, a: Any=a) -> int:
        return merge_hashes(acc, hash_1(o))

    return fold_1(folder, 0, a)


__all__ = ["merge_hashes", "hash_date_time", "hash_1", "box_hash_option", "box_hash_array", "box_hash_seq"]




from __future__ import annotations
from collections.abc import Callable
from typing import (Any, TypeVar)
from ..arctrl_core.comment import Comment
from ..fable_library.list import (choose, of_array, FSharpList)
from ..fable_library.option import (map, value as value_6)
from ..fable_library.seq import map as map_1
from ..fable_library.string_ import replace
from ..fable_library.types import to_string
from ..fable_library.util import IEnumerable_1
from ..thoth_json_core.decode import (object, IOptionalGetter, string, IGetters, map as map_2)
from ..thoth_json_core.types import (IEncodable, IEncoderHelpers_1, Decoder_1)
from .context.rocrate.isa_comment_context import context_jsonvalue
from .encode import try_include
from .idtable import encode

__A_ = TypeVar("__A_")

def encoder(comment: Comment) -> IEncodable:
    def chooser(tupled_arg: tuple[str, IEncodable | None], comment: Any=comment) -> tuple[str, IEncodable] | None:
        def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
            return (tupled_arg[0], v_1)

        return map(mapping, tupled_arg[1])

    def _arrow1970(value: str, comment: Any=comment) -> IEncodable:
        class ObjectExpr1969(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                return helpers.encode_string(value)

        return ObjectExpr1969()

    def _arrow1972(value_2: str, comment: Any=comment) -> IEncodable:
        class ObjectExpr1971(IEncodable):
            def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
                return helpers_1.encode_string(value_2)

        return ObjectExpr1971()

    values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([try_include("name", _arrow1970, comment.Name), try_include("value", _arrow1972, comment.Value)]))
    class ObjectExpr1973(IEncodable):
        def Encode(self, helpers_2: IEncoderHelpers_1[Any], comment: Any=comment) -> Any:
            def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_2))

            arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
            return helpers_2.encode_object(arg)

    return ObjectExpr1973()


def _arrow1976(get: IGetters) -> Comment:
    def _arrow1974(__unit: None=None) -> str | None:
        object_arg: IOptionalGetter = get.Optional
        return object_arg.Field("name", string)

    def _arrow1975(__unit: None=None) -> str | None:
        object_arg_1: IOptionalGetter = get.Optional
        return object_arg_1.Field("value", string)

    return Comment(_arrow1974(), _arrow1975())


decoder: Decoder_1[Comment] = object(_arrow1976)

def ROCrate_genID(c: Comment) -> str:
    match_value: str | None = c.Name
    if match_value is None:
        return "#EmptyComment"

    else: 
        n: str = match_value
        v: str = ("_" + replace(value_6(c.Value), " ", "_")) if (c.Value is not None) else ""
        return ("#Comment_" + replace(n, " ", "_")) + v



def ROCrate_encoder(comment: Comment) -> IEncodable:
    def chooser(tupled_arg: tuple[str, IEncodable | None], comment: Any=comment) -> tuple[str, IEncodable] | None:
        def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
            return (tupled_arg[0], v_1)

        return map(mapping, tupled_arg[1])

    def _arrow1980(__unit: None=None, comment: Any=comment) -> IEncodable:
        value: str = ROCrate_genID(comment)
        class ObjectExpr1979(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                return helpers.encode_string(value)

        return ObjectExpr1979()

    class ObjectExpr1981(IEncodable):
        def Encode(self, helpers_1: IEncoderHelpers_1[Any], comment: Any=comment) -> Any:
            return helpers_1.encode_string("Comment")

    def _arrow1983(value_2: str, comment: Any=comment) -> IEncodable:
        class ObjectExpr1982(IEncodable):
            def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
                return helpers_2.encode_string(value_2)

        return ObjectExpr1982()

    def _arrow1985(value_4: str, comment: Any=comment) -> IEncodable:
        class ObjectExpr1984(IEncodable):
            def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
                return helpers_3.encode_string(value_4)

        return ObjectExpr1984()

    values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([("@id", _arrow1980()), ("@type", ObjectExpr1981()), try_include("name", _arrow1983, comment.Name), try_include("value", _arrow1985, comment.Value), ("@context", context_jsonvalue)]))
    class ObjectExpr1986(IEncodable):
        def Encode(self, helpers_4: IEncoderHelpers_1[Any], comment: Any=comment) -> Any:
            def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_4))

            arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
            return helpers_4.encode_object(arg)

    return ObjectExpr1986()


def _arrow1989(get: IGetters) -> Comment:
    def _arrow1987(__unit: None=None) -> str | None:
        object_arg: IOptionalGetter = get.Optional
        return object_arg.Field("name", string)

    def _arrow1988(__unit: None=None) -> str | None:
        object_arg_1: IOptionalGetter = get.Optional
        return object_arg_1.Field("value", string)

    return Comment(_arrow1987(), _arrow1988())


ROCrate_decoder: Decoder_1[Comment] = object(_arrow1989)

def ROCrate_encoderDisambiguatingDescription(comment: Comment) -> IEncodable:
    value: str = to_string(comment)
    class ObjectExpr1990(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any], comment: Any=comment) -> Any:
            return helpers.encode_string(value)

    return ObjectExpr1990()


def ctor(s: str) -> Comment:
    return Comment.from_string(s)


ROCrate_decoderDisambiguatingDescription: Decoder_1[Comment] = map_2(ctor, string)

def ISAJson_encoder(id_map: Any | None, comment: Comment) -> IEncodable:
    def f(comment_1: Comment, id_map: Any=id_map, comment: Any=comment) -> IEncodable:
        def chooser(tupled_arg: tuple[str, IEncodable | None], comment_1: Any=comment_1) -> tuple[str, IEncodable] | None:
            def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
                return (tupled_arg[0], v_1)

            return map(mapping, tupled_arg[1])

        def _arrow1994(value: str, comment_1: Any=comment_1) -> IEncodable:
            class ObjectExpr1993(IEncodable):
                def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                    return helpers.encode_string(value)

            return ObjectExpr1993()

        def _arrow1996(value_2: str, comment_1: Any=comment_1) -> IEncodable:
            class ObjectExpr1995(IEncodable):
                def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
                    return helpers_1.encode_string(value_2)

            return ObjectExpr1995()

        def _arrow1998(value_4: str, comment_1: Any=comment_1) -> IEncodable:
            class ObjectExpr1997(IEncodable):
                def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
                    return helpers_2.encode_string(value_4)

            return ObjectExpr1997()

        values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([try_include("@id", _arrow1994, ROCrate_genID(comment_1)), try_include("name", _arrow1996, comment_1.Name), try_include("value", _arrow1998, comment_1.Value)]))
        class ObjectExpr1999(IEncodable):
            def Encode(self, helpers_3: IEncoderHelpers_1[Any], comment_1: Any=comment_1) -> Any:
                def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                    return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_3))

                arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
                return helpers_3.encode_object(arg)

        return ObjectExpr1999()

    if id_map is None:
        return f(comment)

    else: 
        def _arrow2000(c: Comment, id_map: Any=id_map, comment: Any=comment) -> str:
            return ROCrate_genID(c)

        return encode(_arrow2000, f, comment, id_map)



ISAJson_decoder: Decoder_1[Comment] = decoder

__all__ = ["encoder", "decoder", "ROCrate_genID", "ROCrate_encoder", "ROCrate_decoder", "ROCrate_encoderDisambiguatingDescription", "ROCrate_decoderDisambiguatingDescription", "ISAJson_encoder", "ISAJson_decoder"]




from abc import abstractmethod
from typing import Protocol

class IISAPrintable(Protocol):
    @abstractmethod
    def Print(self) -> str:
        ...

    @abstractmethod
    def PrintCompact(self) -> str:
        ...





from __future__ import annotations
from collections.abc import Callable
from typing import (Any, TypeVar)
from ..arctrl_core.comment import Comment
from ..arctrl_core.ontology_annotation import OntologyAnnotation
from ..arctrl_core.Process.column_index import order_name
from ..arctrl_core.uri import URIModule_toString
from ..fable_library.list import (of_array, choose, FSharpList)
from ..fable_library.option import map as map_1
from ..fable_library.seq import (map as map_2, filter)
from ..fable_library.string_ import replace
from ..fable_library.types import (to_string, Array)
from ..fable_library.util import (int32_to_string, IEnumerable_1)
from ..thoth_json_core.decode import (one_of, map, int_1, float_1, string, object, IOptionalGetter, resize_array, IGetters)
from ..thoth_json_core.types import (Decoder_1, IEncodable, IEncoderHelpers_1)
from .comment import (encoder, decoder, ROCrate_encoderDisambiguatingDescription, ROCrate_decoderDisambiguatingDescription)
from .context.rocrate.isa_ontology_annotation_context import context_jsonvalue
from .context.rocrate.property_value_context import context_jsonvalue as context_jsonvalue_1
from .encode import (try_include, try_include_seq)
from .idtable import encode
from .string_table import (encode_string, decode_string)

__A_ = TypeVar("__A_")

AnnotationValue_decoder: Decoder_1[str] = one_of(of_array([map(int32_to_string, int_1), map(to_string, float_1), string]))

def OntologyAnnotation_encoder(oa: OntologyAnnotation) -> IEncodable:
    def chooser(tupled_arg: tuple[str, IEncodable | None], oa: Any=oa) -> tuple[str, IEncodable] | None:
        def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
            return (tupled_arg[0], v_1)

        return map_1(mapping, tupled_arg[1])

    def _arrow2004(value: str, oa: Any=oa) -> IEncodable:
        class ObjectExpr2003(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                return helpers.encode_string(value)

        return ObjectExpr2003()

    def _arrow2006(value_2: str, oa: Any=oa) -> IEncodable:
        class ObjectExpr2005(IEncodable):
            def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
                return helpers_1.encode_string(value_2)

        return ObjectExpr2005()

    def _arrow2008(value_4: str, oa: Any=oa) -> IEncodable:
        class ObjectExpr2007(IEncodable):
            def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
                return helpers_2.encode_string(value_4)

        return ObjectExpr2007()

    def _arrow2009(comment: Comment, oa: Any=oa) -> IEncodable:
        return encoder(comment)

    values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([try_include("annotationValue", _arrow2004, oa.Name), try_include("termSource", _arrow2006, oa.TermSourceREF), try_include("termAccession", _arrow2008, oa.TermAccessionNumber), try_include_seq("comments", _arrow2009, oa.Comments)]))
    class ObjectExpr2010(IEncodable):
        def Encode(self, helpers_3: IEncoderHelpers_1[Any], oa: Any=oa) -> Any:
            def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_3))

            arg: IEnumerable_1[tuple[str, __A_]] = map_2(mapping_1, values)
            return helpers_3.encode_object(arg)

    return ObjectExpr2010()


def _arrow2015(get: IGetters) -> OntologyAnnotation:
    def _arrow2011(__unit: None=None) -> str | None:
        object_arg: IOptionalGetter = get.Optional
        return object_arg.Field("annotationValue", AnnotationValue_decoder)

    def _arrow2012(__unit: None=None) -> str | None:
        object_arg_1: IOptionalGetter = get.Optional
        return object_arg_1.Field("termSource", string)

    def _arrow2013(__unit: None=None) -> str | None:
        object_arg_2: IOptionalGetter = get.Optional
        return object_arg_2.Field("termAccession", string)

    def _arrow2014(__unit: None=None) -> Array[Comment] | None:
        arg_7: Decoder_1[Array[Comment]] = resize_array(decoder)
        object_arg_3: IOptionalGetter = get.Optional
        return object_arg_3.Field("comments", arg_7)

    return OntologyAnnotation.create(_arrow2011(), _arrow2012(), _arrow2013(), _arrow2014())


OntologyAnnotation_decoder: Decoder_1[OntologyAnnotation] = object(_arrow2015)

def OntologyAnnotation_compressedEncoder(string_table: Any, oa: OntologyAnnotation) -> IEncodable:
    def chooser(tupled_arg: tuple[str, IEncodable | None], string_table: Any=string_table, oa: Any=oa) -> tuple[str, IEncodable] | None:
        def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
            return (tupled_arg[0], v_1)

        return map_1(mapping, tupled_arg[1])

    def _arrow2017(s: str, string_table: Any=string_table, oa: Any=oa) -> IEncodable:
        return encode_string(string_table, s)

    def _arrow2018(s_1: str, string_table: Any=string_table, oa: Any=oa) -> IEncodable:
        return encode_string(string_table, s_1)

    def _arrow2019(s_2: str, string_table: Any=string_table, oa: Any=oa) -> IEncodable:
        return encode_string(string_table, s_2)

    def _arrow2020(comment: Comment, string_table: Any=string_table, oa: Any=oa) -> IEncodable:
        return encoder(comment)

    values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([try_include("a", _arrow2017, oa.Name), try_include("ts", _arrow2018, oa.TermSourceREF), try_include("ta", _arrow2019, oa.TermAccessionNumber), try_include_seq("comments", _arrow2020, oa.Comments)]))
    class ObjectExpr2021(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any], string_table: Any=string_table, oa: Any=oa) -> Any:
            def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers))

            arg: IEnumerable_1[tuple[str, __A_]] = map_2(mapping_1, values)
            return helpers.encode_object(arg)

    return ObjectExpr2021()


def OntologyAnnotation_compressedDecoder(string_table: Array[str]) -> Decoder_1[OntologyAnnotation]:
    def _arrow2026(get: IGetters, string_table: Any=string_table) -> OntologyAnnotation:
        def _arrow2022(__unit: None=None) -> str | None:
            arg_1: Decoder_1[str] = decode_string(string_table)
            object_arg: IOptionalGetter = get.Optional
            return object_arg.Field("a", arg_1)

        def _arrow2023(__unit: None=None) -> str | None:
            arg_3: Decoder_1[str] = decode_string(string_table)
            object_arg_1: IOptionalGetter = get.Optional
            return object_arg_1.Field("ts", arg_3)

        def _arrow2024(__unit: None=None) -> str | None:
            arg_5: Decoder_1[str] = decode_string(string_table)
            object_arg_2: IOptionalGetter = get.Optional
            return object_arg_2.Field("ta", arg_5)

        def _arrow2025(__unit: None=None) -> Array[Comment] | None:
            arg_7: Decoder_1[Array[Comment]] = resize_array(decoder)
            object_arg_3: IOptionalGetter = get.Optional
            return object_arg_3.Field("comments", arg_7)

        return OntologyAnnotation(_arrow2022(), _arrow2023(), _arrow2024(), _arrow2025())

    return object(_arrow2026)


def OntologyAnnotation_ROCrate_genID(o: OntologyAnnotation) -> str:
    match_value: str | None = o.TermAccessionNumber
    if match_value is None:
        match_value_1: str | None = o.TermSourceREF
        if match_value_1 is None:
            match_value_2: str | None = o.Name
            if match_value_2 is None:
                return "#DummyOntologyAnnotation"

            else: 
                return "#UserTerm_" + replace(match_value_2, " ", "_")


        else: 
            return "#" + replace(match_value_1, " ", "_")


    else: 
        return URIModule_toString(match_value)



def OntologyAnnotation_ROCrate_encoderDefinedTerm(oa: OntologyAnnotation) -> IEncodable:
    def chooser(tupled_arg: tuple[str, IEncodable | None], oa: Any=oa) -> tuple[str, IEncodable] | None:
        def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
            return (tupled_arg[0], v_1)

        return map_1(mapping, tupled_arg[1])

    def _arrow2030(__unit: None=None, oa: Any=oa) -> IEncodable:
        value: str = OntologyAnnotation_ROCrate_genID(oa)
        class ObjectExpr2029(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                return helpers.encode_string(value)

        return ObjectExpr2029()

    class ObjectExpr2031(IEncodable):
        def Encode(self, helpers_1: IEncoderHelpers_1[Any], oa: Any=oa) -> Any:
            return helpers_1.encode_string("OntologyAnnotation")

    def _arrow2033(value_2: str, oa: Any=oa) -> IEncodable:
        class ObjectExpr2032(IEncodable):
            def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
                return helpers_2.encode_string(value_2)

        return ObjectExpr2032()

    def _arrow2035(value_4: str, oa: Any=oa) -> IEncodable:
        class ObjectExpr2034(IEncodable):
            def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
                return helpers_3.encode_string(value_4)

        return ObjectExpr2034()

    def _arrow2037(value_6: str, oa: Any=oa) -> IEncodable:
        class ObjectExpr2036(IEncodable):
            def Encode(self, helpers_4: IEncoderHelpers_1[Any]) -> Any:
                return helpers_4.encode_string(value_6)

        return ObjectExpr2036()

    def _arrow2038(comment: Comment, oa: Any=oa) -> IEncodable:
        return ROCrate_encoderDisambiguatingDescription(comment)

    values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([("@id", _arrow2030()), ("@type", ObjectExpr2031()), try_include("annotationValue", _arrow2033, oa.Name), try_include("termSource", _arrow2035, oa.TermSourceREF), try_include("termAccession", _arrow2037, oa.TermAccessionNumber), try_include_seq("comments", _arrow2038, oa.Comments), ("@context", context_jsonvalue)]))
    class ObjectExpr2039(IEncodable):
        def Encode(self, helpers_5: IEncoderHelpers_1[Any], oa: Any=oa) -> Any:
            def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_5))

            arg: IEnumerable_1[tuple[str, __A_]] = map_2(mapping_1, values)
            return helpers_5.encode_object(arg)

    return ObjectExpr2039()


def _arrow2044(get: IGetters) -> OntologyAnnotation:
    def _arrow2040(__unit: None=None) -> str | None:
        object_arg: IOptionalGetter = get.Optional
        return object_arg.Field("annotationValue", AnnotationValue_decoder)

    def _arrow2041(__unit: None=None) -> str | None:
        object_arg_1: IOptionalGetter = get.Optional
        return object_arg_1.Field("termSource", string)

    def _arrow2042(__unit: None=None) -> str | None:
        object_arg_2: IOptionalGetter = get.Optional
        return object_arg_2.Field("termAccession", string)

    def _arrow2043(__unit: None=None) -> Array[Comment] | None:
        arg_7: Decoder_1[Array[Comment]] = resize_array(ROCrate_decoderDisambiguatingDescription)
        object_arg_3: IOptionalGetter = get.Optional
        return object_arg_3.Field("comments", arg_7)

    return OntologyAnnotation.create(_arrow2040(), _arrow2041(), _arrow2042(), _arrow2043())


OntologyAnnotation_ROCrate_decoderDefinedTerm: Decoder_1[OntologyAnnotation] = object(_arrow2044)

def OntologyAnnotation_ROCrate_encoderPropertyValue(oa: OntologyAnnotation) -> IEncodable:
    def chooser(tupled_arg: tuple[str, IEncodable | None], oa: Any=oa) -> tuple[str, IEncodable] | None:
        def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
            return (tupled_arg[0], v_1)

        return map_1(mapping, tupled_arg[1])

    def _arrow2048(__unit: None=None, oa: Any=oa) -> IEncodable:
        value: str = OntologyAnnotation_ROCrate_genID(oa)
        class ObjectExpr2047(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                return helpers.encode_string(value)

        return ObjectExpr2047()

    class ObjectExpr2049(IEncodable):
        def Encode(self, helpers_1: IEncoderHelpers_1[Any], oa: Any=oa) -> Any:
            return helpers_1.encode_string("PropertyValue")

    def _arrow2051(value_2: str, oa: Any=oa) -> IEncodable:
        class ObjectExpr2050(IEncodable):
            def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
                return helpers_2.encode_string(value_2)

        return ObjectExpr2050()

    def _arrow2053(value_4: str, oa: Any=oa) -> IEncodable:
        class ObjectExpr2052(IEncodable):
            def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
                return helpers_3.encode_string(value_4)

        return ObjectExpr2052()

    def _arrow2054(comment: Comment, oa: Any=oa) -> IEncodable:
        return ROCrate_encoderDisambiguatingDescription(comment)

    values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([("@id", _arrow2048()), ("@type", ObjectExpr2049()), try_include("category", _arrow2051, oa.Name), try_include("categoryCode", _arrow2053, oa.TermAccessionNumber), try_include_seq("comments", _arrow2054, oa.Comments), ("@context", context_jsonvalue_1)]))
    class ObjectExpr2055(IEncodable):
        def Encode(self, helpers_4: IEncoderHelpers_1[Any], oa: Any=oa) -> Any:
            def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_4))

            arg: IEnumerable_1[tuple[str, __A_]] = map_2(mapping_1, values)
            return helpers_4.encode_object(arg)

    return ObjectExpr2055()


def _arrow2059(get: IGetters) -> OntologyAnnotation:
    def _arrow2056(__unit: None=None) -> str | None:
        object_arg: IOptionalGetter = get.Optional
        return object_arg.Field("category", string)

    def _arrow2057(__unit: None=None) -> str | None:
        object_arg_1: IOptionalGetter = get.Optional
        return object_arg_1.Field("categoryCode", string)

    def _arrow2058(__unit: None=None) -> Array[Comment] | None:
        arg_5: Decoder_1[Array[Comment]] = resize_array(ROCrate_decoderDisambiguatingDescription)
        object_arg_2: IOptionalGetter = get.Optional
        return object_arg_2.Field("comments", arg_5)

    return OntologyAnnotation.create(_arrow2056(), None, _arrow2057(), _arrow2058())


OntologyAnnotation_ROCrate_decoderPropertyValue: Decoder_1[OntologyAnnotation] = object(_arrow2059)

def OntologyAnnotation_ISAJson_encoder(id_map: Any | None, oa: OntologyAnnotation) -> IEncodable:
    def f(oa_1: OntologyAnnotation, id_map: Any=id_map, oa: Any=oa) -> IEncodable:
        def predicate(c: Comment, oa_1: Any=oa_1) -> bool:
            match_value: str | None = c.Name
            (pattern_matching_result,) = (None,)
            if match_value is not None:
                if match_value == order_name:
                    pattern_matching_result = 0

                else: 
                    pattern_matching_result = 1


            else: 
                pattern_matching_result = 1

            if pattern_matching_result == 0:
                return False

            elif pattern_matching_result == 1:
                return True


        comments: IEnumerable_1[Comment] = filter(predicate, oa_1.Comments)
        def chooser(tupled_arg: tuple[str, IEncodable | None], oa_1: Any=oa_1) -> tuple[str, IEncodable] | None:
            def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
                return (tupled_arg[0], v_1)

            return map_1(mapping, tupled_arg[1])

        def _arrow2063(value: str, oa_1: Any=oa_1) -> IEncodable:
            class ObjectExpr2062(IEncodable):
                def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                    return helpers.encode_string(value)

            return ObjectExpr2062()

        def _arrow2065(value_2: str, oa_1: Any=oa_1) -> IEncodable:
            class ObjectExpr2064(IEncodable):
                def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
                    return helpers_1.encode_string(value_2)

            return ObjectExpr2064()

        def _arrow2067(value_4: str, oa_1: Any=oa_1) -> IEncodable:
            class ObjectExpr2066(IEncodable):
                def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
                    return helpers_2.encode_string(value_4)

            return ObjectExpr2066()

        def _arrow2069(value_6: str, oa_1: Any=oa_1) -> IEncodable:
            class ObjectExpr2068(IEncodable):
                def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
                    return helpers_3.encode_string(value_6)

            return ObjectExpr2068()

        def _arrow2070(comment: Comment, oa_1: Any=oa_1) -> IEncodable:
            return encoder(comment)

        values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([try_include("@id", _arrow2063, OntologyAnnotation_ROCrate_genID(oa_1)), try_include("annotationValue", _arrow2065, oa_1.Name), try_include("termSource", _arrow2067, oa_1.TermSourceREF), try_include("termAccession", _arrow2069, oa_1.TermAccessionNumber), try_include_seq("comments", _arrow2070, comments)]))
        class ObjectExpr2071(IEncodable):
            def Encode(self, helpers_4: IEncoderHelpers_1[Any], oa_1: Any=oa_1) -> Any:
                def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                    return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_4))

                arg: IEnumerable_1[tuple[str, __A_]] = map_2(mapping_1, values)
                return helpers_4.encode_object(arg)

        return ObjectExpr2071()

    if id_map is not None:
        def _arrow2072(o: OntologyAnnotation, id_map: Any=id_map, oa: Any=oa) -> str:
            return OntologyAnnotation_ROCrate_genID(o)

        return encode(_arrow2072, f, oa, id_map)

    else: 
        return f(oa)



OntologyAnnotation_ISAJson_decoder: Decoder_1[OntologyAnnotation] = OntologyAnnotation_decoder

__all__ = ["AnnotationValue_decoder", "OntologyAnnotation_encoder", "OntologyAnnotation_decoder", "OntologyAnnotation_compressedEncoder", "OntologyAnnotation_compressedDecoder", "OntologyAnnotation_ROCrate_genID", "OntologyAnnotation_ROCrate_encoderDefinedTerm", "OntologyAnnotation_ROCrate_decoderDefinedTerm", "OntologyAnnotation_ROCrate_encoderPropertyValue", "OntologyAnnotation_ROCrate_decoderPropertyValue", "OntologyAnnotation_ISAJson_encoder", "OntologyAnnotation_ISAJson_decoder"]




from __future__ import annotations
from collections.abc import Callable
from typing import (Any, TypeVar)
from ...arctrl_core.ontology_annotation import OntologyAnnotation
from ...arctrl_core.Process.process_parameter_value import ProcessParameterValue
from ...arctrl_core.Process.protocol_parameter import ProtocolParameter
from ...arctrl_core.value import Value
from ...fable_library.list import (choose, of_array, FSharpList)
from ...fable_library.option import map
from ...fable_library.seq import map as map_1
from ...fable_library.util import IEnumerable_1
from ...thoth_json_core.decode import (object, IOptionalGetter, IGetters)
from ...thoth_json_core.types import (IEncodable, Decoder_1, IEncoderHelpers_1)
from ..encode import try_include
from ..ontology_annotation import (OntologyAnnotation_ISAJson_encoder, OntologyAnnotation_ISAJson_decoder)
from ..property_value import (encoder, decoder)
from .protocol_parameter import (ISAJson_encoder as ISAJson_encoder_1, ISAJson_decoder as ISAJson_decoder_1)
from .value import (encoder as encoder_1, decoder as decoder_1)

__A_ = TypeVar("__A_")

ROCrate_encoder: Callable[[ProcessParameterValue], IEncodable] = encoder

def _arrow2610(alternate_name: str | None=None, measurement_method: str | None=None, description: str | None=None, category: OntologyAnnotation | None=None, value: Value | None=None, unit: OntologyAnnotation | None=None) -> ProcessParameterValue:
    return ProcessParameterValue.create_as_pv(alternate_name, measurement_method, description, category, value, unit)


ROCrate_decoder: Decoder_1[ProcessParameterValue] = decoder(_arrow2610)

def ISAJson_genID(oa: ProcessParameterValue) -> Any:
    raise Exception("Not implemented")


def ISAJson_encoder(id_map: Any | None, oa: ProcessParameterValue) -> IEncodable:
    def chooser(tupled_arg: tuple[str, IEncodable | None], id_map: Any=id_map, oa: Any=oa) -> tuple[str, IEncodable] | None:
        def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
            return (tupled_arg[0], v_1)

        return map(mapping, tupled_arg[1])

    def _arrow2612(value: ProtocolParameter, id_map: Any=id_map, oa: Any=oa) -> IEncodable:
        return ISAJson_encoder_1(id_map, value)

    def _arrow2613(value_1: Value, id_map: Any=id_map, oa: Any=oa) -> IEncodable:
        return encoder_1(id_map, value_1)

    def _arrow2614(oa_1: OntologyAnnotation, id_map: Any=id_map, oa: Any=oa) -> IEncodable:
        return OntologyAnnotation_ISAJson_encoder(id_map, oa_1)

    values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([try_include("category", _arrow2612, oa.Category), try_include("value", _arrow2613, oa.Value), try_include("unit", _arrow2614, oa.Unit)]))
    class ObjectExpr2615(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any], id_map: Any=id_map, oa: Any=oa) -> Any:
            def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers))

            arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
            return helpers.encode_object(arg)

    return ObjectExpr2615()


def _arrow2619(get: IGetters) -> ProcessParameterValue:
    def _arrow2616(__unit: None=None) -> ProtocolParameter | None:
        object_arg: IOptionalGetter = get.Optional
        return object_arg.Field("category", ISAJson_decoder_1)

    def _arrow2617(__unit: None=None) -> Value | None:
        object_arg_1: IOptionalGetter = get.Optional
        return object_arg_1.Field("value", decoder_1)

    def _arrow2618(__unit: None=None) -> OntologyAnnotation | None:
        object_arg_2: IOptionalGetter = get.Optional
        return object_arg_2.Field("unit", OntologyAnnotation_ISAJson_decoder)

    return ProcessParameterValue(_arrow2616(), _arrow2617(), _arrow2618())


ISAJson_decoder: Decoder_1[ProcessParameterValue] = object(_arrow2619)

__all__ = ["ROCrate_encoder", "ROCrate_decoder", "ISAJson_genID", "ISAJson_encoder", "ISAJson_decoder"]




from ..fable_library.option import default_arg

DefaultWhitespace: int = 2

def default_whitespace(spaces: int | None=None) -> int:
    return default_arg(spaces, DefaultWhitespace)


__all__ = ["DefaultWhitespace", "default_whitespace"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ...dynamic_obj.dynamic_obj import DynamicObj
from ...dynamic_obj.dyn_obj import (try_get_property_value, set_property, set_optional_property)
from ...fable_library.option import value as value_1
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.types import FSharpRef
from .dataset import (Dataset, Dataset_reflection)

def _expr1543() -> TypeInfo:
    return class_type("ARCtrl.ROCrate.Investigation", None, Investigation, Dataset_reflection())


class Investigation(Dataset):
    def __init__(self, id: str, identifier: str, citation: Any | None=None, comment: Any | None=None, creator: Any | None=None, date_created: Any | None=None, date_modified: Any | None=None, date_published: Any | None=None, has_part: Any | None=None, headline: Any | None=None, mentions: Any | None=None, url: Any | None=None, description: Any | None=None) -> None:
        super().__init__(id, ["Investigation"])
        this: FSharpRef[Investigation] = FSharpRef(None)
        this.contents = self
        self.init_00408_1: int = 1
        set_property("identifier", identifier, this.contents)
        set_optional_property("citation", citation, this.contents)
        set_optional_property("comment", comment, this.contents)
        set_optional_property("creator", creator, this.contents)
        set_optional_property("dateCreated", date_created, this.contents)
        set_optional_property("dateModified", date_modified, this.contents)
        set_optional_property("datePublished", date_published, this.contents)
        set_optional_property("hasPart", has_part, this.contents)
        set_optional_property("headline", headline, this.contents)
        set_optional_property("mentions", mentions, this.contents)
        set_optional_property("url", url, this.contents)
        set_optional_property("description", description, this.contents)

    def GetIdentifier(self, __unit: None=None) -> str:
        this: Investigation = self
        obj: DynamicObj = this
        if try_get_property_value("identifier", obj) is not None:
            match_value: str | None
            match_value_1: Any | None = obj.TryGetPropertyValue("identifier")
            if match_value_1 is not None:
                o: Any = value_1(match_value_1)
                match_value = o if (str(type(o)) == "<class \'str\'>") else None

            else: 
                match_value = None

            if match_value is None:
                raise Exception(((((("Property \'" + "identifier") + "\' is set on this \'") + "Investigation") + "\' object but cannot be cast to \'") + "String") + "\'")

            else: 
                return match_value


        else: 
            raise Exception(((("No property \'" + "identifier") + "\' set on this \'") + "Investigation") + "\' object although it is mandatory. Was it created correctly?")


    @staticmethod
    def get_identifier() -> Callable[[Investigation], str]:
        def _arrow1542(inv: Investigation) -> str:
            return inv.GetIdentifier()

        return _arrow1542


Investigation_reflection = _expr1543

def Investigation__ctor_Z47833D48(id: str, identifier: str, citation: Any | None=None, comment: Any | None=None, creator: Any | None=None, date_created: Any | None=None, date_modified: Any | None=None, date_published: Any | None=None, has_part: Any | None=None, headline: Any | None=None, mentions: Any | None=None, url: Any | None=None, description: Any | None=None) -> Investigation:
    return Investigation(id, identifier, citation, comment, creator, date_created, date_modified, date_published, has_part, headline, mentions, url, description)


__all__ = ["Investigation_reflection"]




from __future__ import annotations
from abc import abstractmethod
from typing import Protocol
from .ontology_annotation import OntologyAnnotation
from .value import Value

class IPropertyValue(Protocol):
    @abstractmethod
    def AlternateName(self) -> str | None:
        ...

    @abstractmethod
    def Description(self) -> str | None:
        ...

    @abstractmethod
    def GetAdditionalType(self) -> str:
        ...

    @abstractmethod
    def GetCategory(self) -> OntologyAnnotation | None:
        ...

    @abstractmethod
    def GetUnit(self) -> OntologyAnnotation | None:
        ...

    @abstractmethod
    def GetValue(self) -> Value | None:
        ...

    @abstractmethod
    def MeasurementMethod(self) -> str | None:
        ...





from __future__ import annotations
from typing import (Any, TypeVar)
from ...fable_library.seq import map
from ...fable_library.util import (to_enumerable, IEnumerable_1)
from ...thoth_json_core.decode import (object, seq as seq_1, IRequiredGetter, IGetters)
from ...thoth_json_core.encode import seq
from ...thoth_json_core.types import (IEncodable, IEncoderHelpers_1, Decoder_1)
from ..fs_workbook import FsWorkbook
from ..fs_worksheet import FsWorksheet
from .worksheet import (encode_rows as encode_rows_1, decode_rows as decode_rows_1, encode_columns as encode_columns_1, decode_columns as decode_columns_1)

__A_ = TypeVar("__A_")

def encode_rows(no_numbering: bool, wb: FsWorkbook) -> IEncodable:
    def mapping(sheet: FsWorksheet, no_numbering: Any=no_numbering, wb: Any=wb) -> IEncodable:
        return encode_rows_1(no_numbering, sheet)

    values: IEnumerable_1[tuple[str, IEncodable]] = to_enumerable([("sheets", seq(map(mapping, wb.GetWorksheets())))])
    class ObjectExpr360(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any], no_numbering: Any=no_numbering, wb: Any=wb) -> Any:
            def mapping_1(tupled_arg: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg[0], tupled_arg[1].Encode(helpers))

            arg: IEnumerable_1[tuple[str, __A_]] = map(mapping_1, values)
            return helpers.encode_object(arg)

    return ObjectExpr360()


def _arrow361(builder: IGetters) -> FsWorkbook:
    wb: FsWorkbook = FsWorkbook()
    ws: IEnumerable_1[FsWorksheet]
    arg_1: Decoder_1[IEnumerable_1[FsWorksheet]] = seq_1(decode_rows_1)
    object_arg: IRequiredGetter = builder.Required
    ws = object_arg.Field("sheets", arg_1)
    wb.AddWorksheets(ws)
    return wb


decode_rows: Decoder_1[FsWorkbook] = object(_arrow361)

def encode_columns(no_numbering: bool, wb: FsWorkbook) -> IEncodable:
    def mapping(sheet: FsWorksheet, no_numbering: Any=no_numbering, wb: Any=wb) -> IEncodable:
        return encode_columns_1(no_numbering, sheet)

    values: IEnumerable_1[tuple[str, IEncodable]] = to_enumerable([("sheets", seq(map(mapping, wb.GetWorksheets())))])
    class ObjectExpr362(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any], no_numbering: Any=no_numbering, wb: Any=wb) -> Any:
            def mapping_1(tupled_arg: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg[0], tupled_arg[1].Encode(helpers))

            arg: IEnumerable_1[tuple[str, __A_]] = map(mapping_1, values)
            return helpers.encode_object(arg)

    return ObjectExpr362()


def _arrow363(builder: IGetters) -> FsWorkbook:
    wb: FsWorkbook = FsWorkbook()
    ws: IEnumerable_1[FsWorksheet]
    arg_1: Decoder_1[IEnumerable_1[FsWorksheet]] = seq_1(decode_columns_1)
    object_arg: IRequiredGetter = builder.Required
    ws = object_arg.Field("sheets", arg_1)
    wb.AddWorksheets(ws)
    return wb


decode_columns: Decoder_1[FsWorkbook] = object(_arrow363)

__all__ = ["encode_rows", "decode_rows", "encode_columns", "decode_columns"]




from __future__ import annotations
from collections.abc import Callable
from dataclasses import dataclass
from typing import Any
from ...fable_library.option import (default_arg, map, bind)
from ...fable_library.reflection import (TypeInfo, string_type, option_type, record_type)
from ...fable_library.types import (to_string, Record, Array)
from ..comment import Comment
from ..Helper.collections_ import Option_fromValueWithDefault
from ..ontology_annotation import (OntologyAnnotation, OntologyAnnotation_reflection)

def _expr751() -> TypeInfo:
    return record_type("ARCtrl.Process.MaterialAttribute", [], MaterialAttribute, lambda: [("ID", option_type(string_type)), ("CharacteristicType", option_type(OntologyAnnotation_reflection()))])


@dataclass(eq = False, repr = False, slots = True)
class MaterialAttribute(Record):
    ID: str | None
    CharacteristicType: OntologyAnnotation | None
    def Print(self, __unit: None=None) -> str:
        this: MaterialAttribute = self
        return to_string(this)

    def PrintCompact(self, __unit: None=None) -> str:
        this: MaterialAttribute = self
        return "OA " + MaterialAttribute__get_NameText(this)


MaterialAttribute_reflection = _expr751

def MaterialAttribute_make(id: str | None=None, characteristic_type: OntologyAnnotation | None=None) -> MaterialAttribute:
    return MaterialAttribute(id, characteristic_type)


def MaterialAttribute_create_A220A8A(Id: str | None=None, CharacteristicType: OntologyAnnotation | None=None) -> MaterialAttribute:
    return MaterialAttribute_make(Id, CharacteristicType)


def MaterialAttribute_get_empty(__unit: None=None) -> MaterialAttribute:
    return MaterialAttribute_create_A220A8A()


def MaterialAttribute_fromString_5980DC03(term: str, source: str, accession: str, comments: Array[Comment] | None=None) -> MaterialAttribute:
    oa: OntologyAnnotation = OntologyAnnotation.create(term, source, accession, comments)
    return MaterialAttribute_make(None, Option_fromValueWithDefault(OntologyAnnotation(), oa))


def MaterialAttribute_toStringObject_Z1E3B85DD(ma: MaterialAttribute) -> dict[str, Any]:
    value: dict[str, Any] = {
        "TermAccessionNumber": "",
        "TermName": "",
        "TermSourceREF": ""
    }
    def mapping(oa: OntologyAnnotation, ma: Any=ma) -> dict[str, Any]:
        return OntologyAnnotation.to_string_object(oa)

    return default_arg(map(mapping, ma.CharacteristicType), value)


def MaterialAttribute__get_NameText(this: MaterialAttribute) -> str:
    def mapping(oa: OntologyAnnotation, this: Any=this) -> str:
        return oa.NameText

    return default_arg(map(mapping, this.CharacteristicType), "")


def MaterialAttribute__get_TryNameText(this: MaterialAttribute) -> str | None:
    def binder(oa: OntologyAnnotation, this: Any=this) -> str | None:
        return oa.Name

    return bind(binder, this.CharacteristicType)


def MaterialAttribute__MapCategory_658CFBF6(this: MaterialAttribute, f: Callable[[OntologyAnnotation], OntologyAnnotation]) -> MaterialAttribute:
    return MaterialAttribute(this.ID, map(f, this.CharacteristicType))


def MaterialAttribute__SetCategory_ZDED3A0F(this: MaterialAttribute, c: OntologyAnnotation) -> MaterialAttribute:
    return MaterialAttribute(this.ID, c)


def MaterialAttribute_tryGetNameText_Z1E3B85DD(ma: MaterialAttribute) -> str:
    return MaterialAttribute__get_NameText(ma)


def MaterialAttribute_getNameText_Z1E3B85DD(ma: MaterialAttribute) -> str | None:
    return MaterialAttribute__get_TryNameText(ma)


def MaterialAttribute_nameEqualsString(name: str, ma: MaterialAttribute) -> bool:
    return MaterialAttribute__get_NameText(ma) == name


__all__ = ["MaterialAttribute_reflection", "MaterialAttribute_make", "MaterialAttribute_create_A220A8A", "MaterialAttribute_get_empty", "MaterialAttribute_fromString_5980DC03", "MaterialAttribute_toStringObject_Z1E3B85DD", "MaterialAttribute__get_NameText", "MaterialAttribute__get_TryNameText", "MaterialAttribute__MapCategory_658CFBF6", "MaterialAttribute__SetCategory_ZDED3A0F", "MaterialAttribute_tryGetNameText_Z1E3B85DD", "MaterialAttribute_getNameText_Z1E3B85DD", "MaterialAttribute_nameEqualsString"]

